
#  $Id: FB.pm,v 1.2 2003/01/04 00:11:03 ryan Exp $
#
#  Copyright (c) 1997,1998,1999 Edmund Mergl
#  Portions Copyright (c) 1994,1995,1996,1997 Tim Bunce
#
#  You may distribute under the terms of either the GNU General Public
#  License or the Artistic License, as specified in the Perl README file.


require 5.005;

$DBD::FB::VERSION = '1.39';

{
    package DBD::FB;

    use DBI ();
    use Exporter ();
    @ISA = qw(DynaLoader Exporter);

    require_version DBI 1.10;

    bootstrap DBD::FB $VERSION;

    $err = 0;		# holds error code   for DBI::err
    $errstr = "";	# holds error string for DBI::errstr
    $drh = undef;	# holds driver handle once initialized

    sub driver{
	return $drh if $drh;
	my($class, $attr) = @_;

	$class .= "::dr";

	# not a 'my' since we use it above to prevent multiple drivers

	($drh) = DBI::_new_drh($class, {
	    'Name' => 'FB',
	    'Version' => $VERSION,
	    'Err'    => \$DBD::FB::err,
	    'Errstr' => \$DBD::FB::errstr,
	    'Attribution' => 'Frontbase DBD by Anjo Krank',
	});

	$drh;
    }
	sub debug
	{
		my ($str) = (@_);
	#	warn(('*' x 20) . $str . ('*' x 20 ). "\n");
	}
    1;
}


{   package DBD::FB::dr; # ====== DRIVER ======
    use strict;
 
    sub data_sources {
        my ($drh, $attr) = @_;
        my ($host, @sources, $sources, $item);
DBD::FB::debug('data_sources');
		$host = $$attr{'fb_host'};
		$host = 'localhost' if(!defined($host));
        my $dbh = DBD::FB::dr::connect($drh, "dbname=;host=$host") or return undef;
        $sources = $dbh->{fb_data_sources};
# 		warn("host >$host< >".join( ",", @$sources)."<\n");
      	foreach $item (@$sources)
      	{
      		$item =~ s/(.*)/dbi:FB:dbname=$1;host=$host/;
      		push(@sources, $item);
      	}
        $dbh->disconnect;
        return @sources;
    }

    sub connect {
        my($drh, $dbname, $user, $auth, $hash)= @_;

        # create a 'blank' dbh
        my $db	 = $dbname;
        my $host = $dbname;
        my $dbpasswd;
        my $flag;
        $db =~ s/.*?dbname\s*=\s*(\w*).*/$1/;
        $host =~ s/^.*host\s*=\s*//;
        $host =~ s/\s*;.*$//;
       	$host = 'localhost' if (!defined($host) || ($host eq $dbname));
       	$user = '' unless defined($user);
        $auth = '' unless defined($auth);
        $dbpasswd = $$hash{'fb_dbpasswd'};

        my($dbh) = DBI::_new_dbh($drh, {
			'User' => $user, 
			'CURRENT_USER' => $user
		});
		$flag = $$hash{'AutoCommit'};
		$flag = 1 unless defined($flag);
		$dbh->{'AutoCommit'} = $flag;
		$flag = $$hash{'RaiseError'};
		$flag = 0 unless defined($flag);
		$dbh->{'RaiseError'} = $flag;
        $dbpasswd = '' unless defined($dbpasswd);
# 		warn("connect:$dbname, $db, $host");
#         # Connect to the database..
        DBD::FB::db::_login($dbh, "$db/$host/$dbpasswd", $user, $auth) or return undef;

        $dbh;
    }
	
	
	sub parse_dsn {
		#if ($DBI::VERSION <= 1.43) {
			my ($class, $dsn) = @_;
			$dsn =~ s/^(dbi):(\w*?)(?:\((.*?)\))?://i or return;
			my ($scheme, $driver, $attr, $attr_hash) = (lc($1), $2, $3);
			$driver ||= $ENV{DBI_DRIVER} || '';
			$attr_hash = { split /\s*=>?\s*|\s*,\s*/, $attr, -1 } if $attr;
			return ($scheme, 'FBmanual', $attr, $attr_hash, $dsn);
		#}
		#else {
			#return DBI->parse_dsn(@_);
		#}
	}
}


{   package DBD::FB::db; # ====== DATABASE ======
	use DBI qw(:sql_types);
    use strict;

    sub prepare {
        my($dbh, $statement, @attribs)= @_;
DBD::FB::debug('prepare');

        # create a 'blank' sth

        my $sth = DBI::_new_sth($dbh, {
            'Statement' => $statement,
        });

        DBD::FB::st::fb_prepare($sth, $statement, @attribs) or return undef;

        $sth;
    }

	sub ping {
		my($dbh) = @_;
		local $dbh->{RaiseError} = 0 if $dbh->{RaiseError};
		my $ret = DBD::FB::db::fb_ping($dbh);
	
		return $ret;
	}
	
	sub fb_digest_password {
		my ($dbh, $password, $username) = @_;
		if(!defined $password) { return undef; }
		if(!defined $username) { $username = $dbh->{Username}; }
		$username =~ s/^\s+|\s+$//g;
		if($username eq '') { $username = '_SYSTEM';}
		return DBD::FB::db::digest_password($password, $username);
	}

    sub table_info {
        my($dbh) = @_;

DBD::FB::debug('table_info');
        my $sth = $dbh->prepare('
			SELECT CURRENT_CATALOG AS "TABLE_CAT", T0."SCHEMA_NAME" AS "TABLE_SCHEM", T1."TABLE_NAME", TABLE_TYPE
			FROM INFORMATION_SCHEMA.SCHEMATA AS T0, INFORMATION_SCHEMA.TABLES AS T1 
			WHERE 
				T0.SCHEMA_PK = T1.SCHEMA_PK AND
				T0."SCHEMA_NAME" <> \'INFORMATION_SCHEMA\';
        ') or return undef;
        $sth->execute or return undef;

        $sth;
    }

    sub tables {
        my($dbh) = @_;
DBD::FB::debug('tables');
        my $sth = $dbh->prepare("
			SELECT \"TABLE_NAME\" 
			FROM INFORMATION_SCHEMA.SCHEMATA AS T0, 
				(INFORMATION_SCHEMA.TABLES LEFT JOIN INFORMATION_SCHEMA.TABLE_CACHE_SETTINGS USING(SCHEMA_PK, TABLE_PK)) AS T1, 
				INFORMATION_SCHEMA.USERS T2 
			WHERE 
				T0.SCHEMA_PK = T1.SCHEMA_PK AND 
				T0.SCHEMA_OWNER_PK = T2.USER_PK AND 
				T2.USER_NAME = CURRENT_USER AND 
				TABLE_TYPE = 'BASE TABLE' 
			ORDER BY \"TABLE_NAME\"
		") or return undef;
        $sth->execute or return undef;
        my (@tables, @relname);
#       warn($sth->rows);
        while (@relname = $sth->fetchrow_array) {
            push @tables, $relname[0];
        }
        $sth->finish;

        return @tables;
    }
#cr137 : updated for latest DBI 1.54
    sub type_info_all {
        my ($dbh) = @_;
	my $names = {
          TYPE_NAME				=> 0,
          DATA_TYPE				=> 1,
          COLUMN_SIZE			=> 2,
          LITERAL_PREFIX		=> 3,
          LITERAL_SUFFIX		=> 4,
          CREATE_PARAMS			=> 5,
          NULLABLE				=> 6,
          CASE_SENSITIVE		=> 7,
          SEARCHABLE			=> 8,
          UNSIGNED_ATTRIBUTE	=> 9,
          FIXED_PREC_SCALE		=> 10,
          AUTO_UNIQUE_VALUE		=> 11,
          LOCAL_TYPE_NAME		=> 12,
          MINIMUM_SCALE			=> 13,
          MAXIMUM_SCALE			=> 14,
		  SQL_DATA_TYPE			=> 15,
		  SQL_DATETIME_SUB		=> 16,
		  NUM_PREC_RADIX		=> 17,
		  INTERVAL_PRECISION	=> 18
        };
		
		my $TWOGIG	= 2147483647;
		my $P		= 'precision';
		my $PS		= 'precision, scale';
		my $LEN		= 'max length';
		my $UN		= undef;
		
DBD::FB::debug('type_info_all');
	my $ti = [
	  $names,
# name							type								size		prefix			suffix					create	null case se unsign	
#fixed	auto	local				min	max	sql_type			sub		radix	interval_prec

[ 'tinyint',					SQL_TINYINT,						3,			$UN,			$UN,					$UN,	1,0,2,0,	
0,	0,'TINYINT',					$UN,$UN,SQL_TINYINT,		$UN,	10, $UN],
[ 'longint',					SQL_BIGINT,							19,			$UN,			$UN,					$UN,	1,0,2,0,
0,	0,'LONGINT',					$UN,$UN,SQL_BIGINT,			$UN,	10,	$UN],
[ 'blob',						SQL_LONGVARBINARY,					$TWOGIG,	$UN,			$UN,					$UN,	1,0,0,$UN,
0,	0,'BLOB',						$UN,$UN,SQL_LONGVARBINARY,	$UN,	$UN,$UN],
[ 'bit varying',				SQL_VARBINARY,						$TWOGIG,	$UN,			$UN,					$LEN,	1,0,2,$UN,	
0,	0,'BIT VARYING',				$UN,$UN,SQL_VARBINARY,		$UN,	$UN,$UN],
[ 'byte',						SQL_BINARY,							$TWOGIG-1,	$UN,			$UN,					$LEN,	1,0,2,$UN,	
0,	0,'BYTE',						$UN,$UN,SQL_BINARY,			$UN,	$UN,$UN],
[ 'bit',						SQL_BINARY,							$TWOGIG-1,	$UN,			$UN,					$LEN,	1,0,2,$UN,	
0,	0,'BIT',						$UN,$UN,SQL_BINARY,			$UN,	$UN,$UN],
[ 'clob',						SQL_LONGVARCHAR,					$TWOGIG,	$UN,			$UN,					$UN,	1,1,0,$UN,	
0,	0,'CLOB',						$UN,$UN,SQL_LONGVARCHAR,	$UN,	$UN,$UN],
[ 'character',					SQL_CHAR,							$TWOGIG-1,    "'",			"'",					$LEN,	1,1,3,$UN,	
0,	0,'CHARACTER',					$UN,$UN,SQL_CHAR,			$UN,	$UN,$UN],
[ 'numeric',					SQL_NUMERIC,						15,			$UN,			$UN,					$PS,	1,0,2,0,	
0,	0,'NUMERIC',					0,	38,	SQL_NUMERIC,		$UN,	10, $UN],
[ 'decimal',					SQL_DECIMAL,						15,			$UN,			$UN,					$PS,	1,0,2,0,	
0,	0,'DECIMAL',					0,	38,	SQL_DECIMAL,		$UN,	10, $UN],
[ 'integer',					SQL_INTEGER,						10,			$UN,			$UN,					$UN,	1,0,2,0,	
0,	0,'INTEGER',					$UN,$UN,SQL_INTEGER,		$UN,	10, $UN],
[ 'smallint',					SQL_SMALLINT,						5,			$UN,			$UN,					$UN,	1,0,2,0,	
0,	0,'SMALLINT',					$UN,$UN,SQL_SMALLINT,		$UN,	10, $UN],
[ 'float',						SQL_FLOAT,							15,			$UN,			$UN,					$P,		1,0,2,0,	
0,	0,'FLOAT',						$UN,$UN,SQL_FLOAT,			$UN,	10, $UN],
[ 'real',						SQL_REAL,							15,			$UN,			$UN,					$UN,	1,0,2,0,	
0,	0,'REAL',						$UN,$UN,SQL_REAL,			$UN,	10, $UN],
[ 'double',						SQL_DOUBLE,							15,			$UN,			$UN,					$UN,	1,0,2,0,	
0,	0,'DOUBLE PRECISION',			$UN,$UN,SQL_DOUBLE,			$UN,	10,	$UN],
[ 'varchar',					SQL_VARCHAR,						$TWOGIG-1,	"'",			"'",					$LEN,	1,1,3,$UN,	
0,	0,'VARCHAR',					$UN,$UN,SQL_VARCHAR,		$UN,	$UN,$UN],
[ 'date',						SQL_TYPE_DATE,						10,			"DATE '",		"'",					$UN,	1,0,2,$UN,	
0,	0,'DATE',						$UN,$UN,SQL_DATETIME,		1,		$UN,$UN],
[ 'time',						SQL_TYPE_TIME,						16,			"TIME '",		"'",					$UN,	1,0,2,$UN,	
0,	0,'TIME',						$UN,$UN,SQL_DATETIME,		2,		$UN,$UN],
[ 'timestamp',					SQL_TYPE_TIMESTAMP,					26,			"TIMESTAMP '",	"'",					$UN,	1,0,2,$UN,	
0,	0,'TIMESTAMP',					$UN,$UN,SQL_DATETIME,		3,		$UN,$UN],
[ 'time with time zone',		SQL_TYPE_TIME_WITH_TIMEZONE,		22,			"TIME '",		"'",					$UN,	1,0,2,$UN,	
0,	0,'TIME WITH TIME ZONE',		$UN,$UN,SQL_DATETIME,		4,		$UN,$UN],
[ 'timestamp with time zone',	SQL_TYPE_TIMESTAMP_WITH_TIMEZONE,	32,			"TIMESTAMP '",	"'",					$UN,	1,0,2,$UN,	
0,	0,'TIMESTAMP WITH TIME ZONE',	$UN,$UN,SQL_DATETIME,		5,		$UN,$UN],
[ 'boolean',					SQL_BOOLEAN,						1,			$UN,			$UN,					$UN,	1,0,2,$UN,	
$UN,0,'BOOLEAN',					$UN,$UN,SQL_BOOLEAN,		$UN,	$UN,$UN],
[ 'interval year',				SQL_INTERVAL_YEAR,					10,			"INTERVAL '",	"' YEAR",				$UN,	1,0,2,0,	
0,	0,'INTERVAL YEAR',				$UN,$UN,SQL_INTERVAL,		1,		$UN,$UN],
[ 'interval month',				SQL_INTERVAL_MONTH,					10,			"INTERVAL '",	"' MONTH",				$UN,	1,0,2,0,	
0,	0,'INTERVAL MONTH',				$UN,$UN,SQL_INTERVAL,		2,		$UN,$UN],
[ 'interval day',				SQL_INTERVAL_DAY,					15,			"INTERVAL '",	"' DAY",				$UN,	1,0,2,0,	
0,	0,'INTERVAL DAY',				$UN,$UN,SQL_INTERVAL,		3,		$UN,$UN],
[ 'interval hour',				SQL_INTERVAL_HOUR,					15,			"INTERVAL '",	"' HOUR",				$UN,	1,0,2,0,	
0,	0,'INTERVAL HOUR',				$UN,$UN,SQL_INTERVAL,		4,		$UN,$UN],
[ 'interval minute',			SQL_INTERVAL_MINUTE,				15,			"INTERVAL '",	"' MINUTE",				$UN,	1,0,2,0,	
0,	0,'INTERVAL MINUTE',			$UN,$UN,SQL_INTERVAL,		5,		$UN,$UN],
[ 'interval second',			SQL_INTERVAL_SECOND,				15,			"INTERVAL '",	"' SECOND",				$UN,	1,0,2,0,	
0,	0,'INTERVAL SECOND',			$UN,$UN,SQL_INTERVAL,		6,		$UN,$UN],
[ 'interval year to month',		SQL_INTERVAL_YEAR_TO_MONTH,			10,			"INTERVAL '",	"' YEAR TO MONTH",		$UN,	1,0,2,0,	
0,	0,'INTERVAL YEAR TO MONTH',		$UN,$UN,SQL_INTERVAL,		7,		$UN,$UN],
[ 'interval day to hour',		SQL_INTERVAL_DAY_TO_HOUR,			15,			"INTERVAL '",	"' DAY TO HOUR",		$UN,	1,0,2,0,	
0,	0,'INTERVAL DAY TO HOUR',		$UN,$UN,SQL_INTERVAL,		8,		$UN,$UN],
[ 'interval day to minute',		SQL_INTERVAL_DAY_TO_MINUTE,			15,			"INTERVAL '",	"' DAY TO MINUTE",		$UN,	1,0,2,0,	
0,	0,'INTERVAL DAY TO MINUTE',		$UN,$UN,SQL_INTERVAL,		9,		$UN,$UN],
[ 'interval day to second',		SQL_INTERVAL_DAY_TO_SECOND,			15,			"INTERVAL '",	"' DAY TO SECOND",		$UN,	1,0,2,0,	
0,	0,'INTERVAL DAY TO SECOND',		$UN,$UN,SQL_INTERVAL,		10,		$UN,$UN],
[ 'interval hour to minute',	SQL_INTERVAL_HOUR_TO_MINUTE,		15,			"INTERVAL '",	"' HOUR TO MINUTE",		$UN,	1,0,2,0,	
0,	0,'INTERVAL HOUR TO MINUTE',	$UN,$UN,SQL_INTERVAL,		11,		$UN,$UN],
[ 'interval hour to second',	SQL_INTERVAL_HOUR_TO_SECOND,		15,			"INTERVAL '",	"' HOUR TO SECOND",		$UN,	1,0,2,0,	
0,	0,'INTERVAL HOUR TO SECOND',	$UN,$UN,SQL_INTERVAL,		12,		$UN,$UN],
[ 'interval minute to second',	SQL_INTERVAL_MINUTE_TO_SECOND,		15,			"INTERVAL '",	"' MINUTE TO SECOND",	$UN,	1,0,2,0,	
0,	0,'INTERVAL MINUTE TO SECOND',	$UN,$UN,SQL_INTERVAL,		13,		$UN,$UN]
# the column size of the INTERVAL YEAR TO MONTH and INTERVAL DAY TO SECOND types indicate these are returned as long integer (months) and double values (seconds).
        ];
	return $ti;
    }

    sub quote {
	my ($dbh, $str, $data_type) = @_;

        return "NULL" unless defined $str;
       
        $str =~ s/\\/\\\\/g;
        $str=~s/'/''/g;

        return "'$str'";
    }
	
	# The get_info function was automatically generated by
	# DBI::DBD::Metadata::write_getinfo_pm v2.008696.

    sub get_info {
        my($dbh, $info_type) = @_;
        require DBD::FB::GetInfo;
        my $v = $DBD::FB::GetInfo::info{int($info_type)};
        $v = $v->($dbh) if ref $v eq 'CODE';
        return $v;
    }

    
#      sub do {
#        my($dbh, $statement )= @_;
# DBD::FB::debug('do');
# 
#         # create a 'blank' sth
# 
#         my $sth = DBI::_new_sth($dbh, {
#             'Statement' => $statement,
#         });
# 
#         DBD::FB::st::_prepare($sth, $statement, ()) or return undef;
# 		$sth->execute();
# 		$sth;
#     }
}


{   package DBD::FB::st; # ====== STATEMENT ======

    # all done in XS
}

1;

__END__


=head1 NAME

DBD::FB - Frontbase database driver for the DBI module

=head1 VERSION

This documents version 1.38 of the DBD::FB module

=head1 SYNOPSIS

  use DBI;

  $dbh = DBI->connect("dbi:FB:dbname=$dbname;host=$host", $user, $passwd);

  # See the DBI module documentation for full details

=head1 DESCRIPTION

DBD::FB is a Perl module which works with the DBI module to provide access to 
Frontbase databases.

=head1 MODULE DOCUMENTATION

This documentation describes driver specific behavior and restrictions. It is 
not supposed to be used as the only reference for the user. In any case 
consult the L<DBI|DBI> documentation first!

=head1 THE DBI CLASS

=head2 DBI Class Methods

=over 4

=item B<parse_dsn>

  ($scheme, $driver, $attr_string, $attr_hash, $driver_dsn) = DBI->parse_dsn($dsn)
      or die "Can't parse DBI DSN '$dsn'";

Implemented by DBI, no driver-specific impact.

=item B<connect>

To connect to a database with a minimum of parameters, use the 
following syntax: 

   $dbh = DBI->connect("dbi:FB:dbname=$dbname", $user, $passwd);

This connects to the database $dbname at localhost as $user using $passwd to 
authenticate. The $user parameter can not be an empty string but the $passwd 
parameter can be empty if required.

If a database password is required then pass this using the driver specific attribute 'fb_dbpasswd'.

The following connect statement shows all FrontBase parameters: 

  $dbh = DBI->connect("dbi:FB:dbname=$dbname;host=$host", $user, $passwd,
                       {fb_dbpasswd => $dbpasswd});

=item B<connect_cached>

Implemented by DBI, no driver-specific impact.

=item B<available_drivers>

  @driver_names = DBI->available_drivers;

Implemented by DBI, no driver-specific impact.

=item B<installed_drivers>

Implemented by DBI, no driver-specific impact.

=item B<installed_versions>

Implemented by DBI, no driver-specific impact.

=item B<data_sources>

  @data_sources = DBI->data_sources('FB');
  @data_sources = DBI->data_sources('FB', {'fb_host => $host'});

The driver supports this method. Use the fb_host attribute 
to specify the available databases for the specified host.
If the fb_host attribute is not used then localhost is the default.

=item B<trace>

  DBI->trace($trace_setting);
  DBI->trace($trace_setting, $trace_filename);
  $trace_setting = DBI->trace;

Implemented by DBI, no driver-specific impact.

=back

=head1 METHODS COMMON TO ALL HANDLES

=over 4

=item B<err>

  $rv = $h->err;

Supported by this driver as proposed by DBI. For the connect 
method a failure returns C<FB_ERR_NO_CONNECTION == 1>. In all other cases it returns 
C<FBresultStatus> of the current handle.

=item B<errstr>

  $str = $h->errstr;

Supported by this driver as proposed by DBI. It returns the 
C<FBerrorMessage> related to the current handle. 

=item B<state>

  $str = $h->state;

The driver does not support this method. This uses the DBI default 
implementation and will return S1000 (General Error) for all errors.

=item B<set_err>

  $rv = $h->set_err($err, $errstr);
  $rv = $h->set_err($err, $errstr, $state);
  $rv = $h->set_err($err, $errstr, $state, $method);
  $rv = $h->set_err($err, $errstr, $state, $method, $rv);

Implemented by DBI, no driver-specific impact.

=item B<trace>

  $h->trace($trace_settings);
  $h->trace($trace_settings, $trace_filename);
  $trace_settings = $h->trace;

Implemented by DBI, no driver-specific impact.

=item B<trace_msg>

  $h->trace_msg($message_text);
  $h->trace_msg($message_text, $min_level);

Implemented by DBI, no driver-specific impact.

=item B<func>
	
  $h->func(@func_arguments, $func_name) or die ...;

Not used by this driver.

=item B<can>

  $is_implemented = $h->can($method_name);

Implemented by DBI, no driver-specific impact.

=item B<parse_trace_flags>

   $trace_settings_integer = $h->parse_trace_flags($trace_settings);
   
Implemented by DBI, no driver-specific impact. The FrontBase driver does not
support any driver specific trace flags.
   
=item B<parse_trace_flag>

	$bit_flag = $h->parse_trace_flag($trace_flag_name);
   
Implemented by DBI, no driver-specific impact. The FrontBase driver does not
support any driver specific trace flags.

=item B<swap_inner_handle>

   $rc = $h1->swap_inner_handle( $h2 );
   $rc = $h1->swap_inner_handle( $h2, $allow_reparent );
   
Implemented by DBI, no driver-specific impact.

=back

=head1 ATTRIBUTES COMMON TO ALL HANDLES

=over 4

=item B<Warn> (boolean, inherited)

Implemented by DBI, no driver-specific impact.

=item B<Active> (boolean, read-only)

Supported by this driver as proposed by DBI. A database 
handle is active while it is connected and  statement 
handle is active until it is finished.

=item B<Executed> (boolean, read-only)

Implemented by DBI, no driver-specific impact. Requires DBI 1.41 or greater.

=item B<Kids> (integer, read-only)

Implemented by DBI, no driver-specific impact.

=item B<ActiveKids> (integer, read-only)

Implemented by DBI, no driver-specific impact.

=item B<CachedKids> (hash ref)

Implemented by DBI, no driver-specific impact.

=item B<Type> (scalar)

Implemented by DBI, no driver-specific impact.

=item B<ChildHandles> (array ref)

Implemented by DBI, no driver-specific impact.

=item B<CompatMode> (boolean, inherited)

Not used by this driver. 

=item B<InactiveDestroy> (boolean)

Implemented by DBI, no driver-specific impact.

=item B<PrintWarn> (boolean, inherited)

Implemented by DBI, no driver-specific impact.

The driver does not currently return warnings so setting this to true has no effect.

=item B<PrintError> (boolean, inherited)

Implemented by DBI, no driver-specific impact.

=item B<RaiseError> (boolean, inherited)

Implemented by DBI, no driver-specific impact.

=item B<HandleError> (code ref, inherited)

Implemented by DBI, no driver-specific impact.

=item B<HandleSetErr> (code ref, inherited)

Implemented by DBI, no driver-specific impact. Requires DBI 1.41 or greater.

=item B<ErrCount> (unsigned integer)

Implemented by DBI, no driver-specific impact. Not affected by any of this driver's methods.

=item B<ShowErrorStatement> (boolean, inherited)

Implemented by DBI, no driver-specific impact.

=item B<TraceLevel> (integer, inherited)

Implemented by DBI, no driver-specific impact.

=item B<FetchHashKeyName> (string, inherited)

Implemented by DBI, no driver-specific impact.

=item B<ChopBlanks> (boolean, inherited)

Supported by this driver as proposed by DBI. This 
method is similar to the SQL-function C<RTRIM>.

This is not being applied to fixed character data.

=item B<LongReadLen> (unsigned integer, inherited)

Not used by this driver.

=item B<LongTruncOk> (boolean, inherited)

Not used by this driver.

=item B<TaintIn> (boolean, inherited)

Implemented by DBI, no driver-specific impact. Requires DBI 1.31 or greater.

=item B<TaintOut> (boolean, inherited)

Implemented by DBI, no driver-specific impact. Requires DBI 1.31 or greater.

=item B<Taint> (boolean, inherited)

Implemented by DBI, no driver-specific impact.

=item B<Profile> (inherited)

Implemented by DBI, no driver-specific impact. Requires DBI 1.24 or greater.

=item B<private_*>

Implemented by DBI, no driver-specific impact.

=back

=head1 DBI DATABASE HANDLE OBJECTS

=head2 Database Handle Methods

=over 4

=item B<clone>

  $new_dbh = $dbh->clone();
  $new_dbh = $dbh->clone(\%attr);

Implemented by DBI, no driver-specific impact. Requires DBI 1.33 or later.

=item B<data_sources>

@ary = $dbh->data_sources();
@ary = $dbh->data_sources(\%attr);

This uses the implementation of DBI->data_sources (see the DBI class methods section). 
This does not support retreiving the list of available databases from the connected host 
unless the fb->host attribute is set to that host. Requires DBI 1.38 or later.

=item B<do>

  $rv  = $dbh->do($statement, \%attr, @bind_values);

Implemented by DBI, no driver-specific impact. See the 
notes for the statement handle's execute method, which 
is in the section on statement handle methods later on
in this document.

=item B<last_insert_id>

  $rv = $dbh->last_insert_id($catalog, $schema, $table, $field);
  $rv = $dbh->last_insert_id($catalog, $schema, $table, $field, \%attr);
  
Not supported by this driver.

=item B<selectrow_array>

  @row_ary = $dbh->selectrow_array($statement);
  @row_ary = $dbh->selectrow_array($statement, \%attr);
  @row_ary = $dbh->selectrow_array($statement, \%attr, @bind_values);

Implemented by DBI, no driver-specific impact.

=item B<selectrow_arrayref>

  $ary_ref = $dbh->selectrow_arrayref($statement);
  $ary_ref = $dbh->selectrow_arrayref($statement, \%attr);
  $ary_ref = $dbh->selectrow_arrayref($statement, \%attr, @bind_values);

Implemented by DBI, no driver-specific impact.

=item B<selectrow_hashref>

  $hash_ref = $dbh->selectrow_hashref($statement);
  $hash_ref = $dbh->selectrow_hashref($statement, \%attr);
  $hash_ref = $dbh->selectrow_hashref($statement, \%attr, @bind_values);

Implemented by DBI, no driver-specific impact.

=item B<selectall_arrayref>

  $ary_ref = $dbh->selectall_arrayref($statement);
  $ary_ref = $dbh->selectall_arrayref($statement, \%attr);
  $ary_ref = $dbh->selectall_arrayref($statement, \%attr, @bind_values);

Implemented by DBI, no driver-specific impact.

=item B<selectall_hashref>

  $hash_ref = $dbh->selectall_hashref($statement, $key_field);

Implemented by DBI, no driver-specific impact.

=item B<selectcol_arrayref>

  $ary_ref = $dbh->selectcol_arrayref($statement);
  $ary_ref = $dbh->selectcol_arrayref($statement, \%attr);
  $ary_ref = $dbh->selectcol_arrayref($statement, \%attr, @bind_values);

Implemented by DBI, no driver-specific impact.

=item B<prepare>

  $sth = $dbh->prepare($statement)          or die $dbh->errstr;
  $sth = $dbh->prepare($statement, \%attr)  or die $dbh->errstr;

The driver always executes a statement directly and does not send 
a statement to the server to be prepared. Hence this method just 
stores the statement after checking for place-holders. No information 
about the statement is available after preparing it.

=item B<prepare_cached>

  $sth = $dbh->prepare_cached($statement)
  $sth = $dbh->prepare_cached($statement, \%attr)
  $sth = $dbh->prepare_cached($statement, \%attr, $if_active)

Implemented by DBI, no driver-specific impact. 
This method is not useful for this driver, because 
preparing a statement has no database interaction.

=item B<commit>

  $rc  = $dbh->commit     or die $dbh->errstr;

Supported by this driver as proposed by DBI. See also the 
notes about B<Transactions> in the FURTHER INFORMATION section below. 

=item B<rollback>

  rc  = $dbh->rollback   or die $dbh->errstr;

Supported by this driver as proposed by DBI. See also the 
notes about B<Transactions> in the FURTHER INFORMATION section below.

=item B<begin_work>

$rc  = $dbh->begin_work   or die $dbh->errstr;

Implemented by DBI, no driver-specific impact. 

The driver does not send an explicit BEGIN to the server as FrontBase automatically
starts a new transaction.

=item B<disconnect>

  $rc = $dbh->disconnect  or warn $dbh->errstr;

Supported by this driver as proposed by DBI. 

=item B<ping>

  $rc = $dbh->ping;

Supported by this driver as proposed by DBI. This returns true if the
handle is connected and undef if not.

=item B<get_info>

$value = $dbh->get_info( $info_type );

Not supported by this driver.

=item B<table_info>

	$sth = $dbh->table_info;

Limited support. This method returns all tables and views. It does not select indices, 
sequences or system tables. It does not support any of the parameter filters and does not return a REMARKS column.

=item B<column_info>

  $sth = $dbh->column_info( $catalog, $schema, $table, $column );
  
Not supported by this driver.

=item B<primary_key_info>

  $sth = $dbh->primary_key_info( $catalog, $schema, $table );
  
Not supported by this driver.

=item B<primary_key>

  @key_column_names = $dbh->primary_key( $catalog, $schema, $table );
  
Not supported by this driver.

=item B<foreign_key_info>

  $sth = $dbh->foreign_key_info( $pk_catalog, $pk_schema, $pk_table
                               , $fk_catalog, $fk_schema, $fk_table );
  $sth = $dbh->foreign_key_info( $pk_catalog, $pk_schema, $pk_table
                               , $fk_catalog, $fk_schema, $fk_table
                               , \%attr );
  
Not supported by this driver.

=item B<statistics_info>

	$sth = $dbh->statistics_info( $catalog, $schema, $table, $unique_only, $quick );

Not supported by this driver.

=item B<tables>

	@names = $dbh->tables( $catalog, $schema, $table, $type );
	@names = $dbh->tables;        # deprecated

Limited support. 

This method returns all tables which are owned by the 
current user. It does not select any views, indices, sequences or 
system tables. It does not filter by the arguments passed.

=item B<type_info_all>

  $type_info_all = $dbh->type_info_all;

Supported by this driver as proposed by DBI.
The information returned is incomplete and out of date for the latest DBI. 
Information is only provided for SQL data-types and for frequently used data-types.

=item B<type_info>

  @type_info = $dbh->type_info($data_type);

Implemented by DBI, no driver-specific impact.
This is currently limited by what type_info_all can return (see above).

=item B<quote>

  $sql = $dbh->quote($value);
  $sql = $dbh->quote($value, $data_type);

This module implements it's own quote method. In addition to the 
DBI method it doubles also the backslash, because Frontbase treats 
a backslash as an escape character. The optional data_type parameter 
is ignored.

=item B<quote_identifier>

	$sql = $dbh->quote_identifier( $name );
	$sql = $dbh->quote_identifier( $catalog, $schema, $table, \%attr );
	
Implemented by DBI, no driver-specific impact.

=item B<take_imp_data>

	$imp_data = $dbh->take_imp_data;
	
Not supported by this driver.

=back

=item B<fb_digest_password>

	$digest = $dbh->func($passwd[, $username], "fb_digest_password");
	$sth = $dbh->prepare('alter user "'.$username.'" set passwd '.$digest."'");
	
This is a module specific function used to generate a password digest from a clear text password and username.
If the username is omitted then the value of $dbh->{Username} is used.

=head2 Database Handle Attributes

=over 4

=item B<AutoCommit>  (boolean)

Supported by this driver as proposed by DBI.

The default of AutoCommit is on. When setting AutoCommit to off, 
a transaction will be started and every commit or rollback 
will automatically start a new transaction, i.e. there is always an 
active transaction. For details see the notes about B<Transactions> elsewhere in this document.

=item B<Driver>  (handle)

Implemented by DBI, no driver-specific impact.

=item B<Name>  (string, read-only)

Not supported by this driver.

=item B<Statement>  (string, read-only)

Implemented by DBI, no driver-specific impact.

=item B<RowCacheSize>  (integer)

Implemented by DBI, not used by this driver.

=item B<Username>  (string, read-only)

Supported by this driver as proposed by DBI.

=item B<fb_auto_escape> (boolean)

Frontbase specific attribute. If true, then quotes and backslashes in all 
parameters will be escaped in the following way: 

  escape quote with a quote (SQL)
  escape backslash with a backslash except for octal presentation.

The default is on. Note, that Frontbase also accepts quotes, which 
are escaped by a backslash. Any other ASCII character can be used 
directly in a string constant. 

=back


=head1 DBI STATEMENT HANDLE OBJECTS

=head2 Statement Handle Methods

=over 4

=item B<bind_param>

  $rv = $sth->bind_param($p_num, $bind_value);
  $rv = $sth->bind_param($p_num, $bind_value, \%attr);
  $rv = $sth->bind_param($p_num, $bind_value, $bind_type);

Supported by this driver as proposed by DBI.
This supports binding FrontBase types using the {fb_type => n} attribute where n is the FrontBase datatype code.
To bind a SQL identifier use an fb_type code of 99.

=item B<bind_param_inout>

Currently not supported by this driver.

=item B<bind_param_array>

  $rc = $sth->bind_param_array($p_num, $array_ref_or_value);
  $rc = $sth->bind_param_array($p_num, $array_ref_or_value, \%attr);
  $rc = $sth->bind_param_array($p_num, $array_ref_or_value, $bind_type);

Supported by this driver as proposed by DBI.

=item B<execute>

  $rv = $sth->execute                or die $sth->errstr;
  $rv = $sth->execute(@bind_values)  or die $sth->errstr;

Supported by this driver as proposed by DBI. 
In addition to 'UPDATE', 'DELETE', 'INSERT' statements, for 
which it returns always the number of affected rows, the execute 
method can also be used for 'SELECT ... INTO table' statements.

=item B<execute_array>

  $tuples = $sth->execute_array(\%attr) or die $sth->errstr;
  $tuples = $sth->execute_array(\%attr, @bind_values) or die $sth->errstr;
  
  ($tuples, $rows) = $sth->execute_array(\%attr) or die $sth->errstr;
  ($tuples, $rows) = $sth->execute_array(\%attr, @bind_values) or die $sth->errstr;
  
Supported by this driver as proposed by DBI.

=item B<execute_for_fetch>

  $tuples = $sth->execute_for_fetch($fetch_tuple_sub);
  $tuples = $sth->execute_for_fetch($fetch_tuple_sub, \@tuple_status);
  
  ($tuples, $rows) = $sth->execute_for_fetch($fetch_tuple_sub);
  ($tuples, $rows) = $sth->execute_for_fetch($fetch_tuple_sub, \@tuple_status);

Supported by this driver as proposed by DBI.

=item B<fetchrow_arrayref>

  $ary_ref = $sth->fetchrow_arrayref;
  $ary_ref = $sth->fetch;    # alias

Supported by this driver as proposed by DBI. 

=item B<fetchrow_array>

  @ary = $sth->fetchrow_array;

Supported by this driver as proposed by DBI. 

=item B<fetchrow_hashref>

  $hash_ref = $sth->fetchrow_hashref;
  $hash_ref = $sth->fetchrow_hashref($name);

Supported by this driver as proposed by DBI. 

=item B<fetchall_arrayref>

  $tbl_ary_ref = $sth->fetchall_arrayref;
  $tbl_ary_ref = $sth->fetchall_arrayref( $slice );
  $tbl_ary_ref = $sth->fetchall_arrayref( $slice, $max_rows  );

Supported by this driver as proposed by DBI.

=item B<fetchall_hashref>

  $hash_ref = $sth->fetchall_hashref($key_field);

Supported by this driver as proposed by DBI.

=item B<finish>

  $rc = $sth->finish;

Supported by this driver as proposed by DBI. 

=item B<rows>

  $rv = $sth->rows;

Supported by this driver as proposed by DBI. 
In contrast to many other drivers the number of rows is 
available immediately after executing the statement.

This returns 0 when the number of rows is unknown. This should return -1.

=item B<bind_col>

  $rc = $sth->bind_col($column_number, \$var_to_bind);
  $rc = $sth->bind_col($column_number, \$var_to_bind, \%attr );
  $rc = $sth->bind_col($column_number, \$var_to_bind, $bind_type );

Supported by this driver as proposed by DBI. 

=item B<bind_columns>

  $rc = $sth->bind_columns(@list_of_refs_to_vars_to_bind);

Supported by this driver as proposed by DBI. 

=item B<dump_results>

  $rows = $sth->dump_results($maxlen, $lsep, $fsep, $fh);

Supported by this driver as proposed by DBI.

=item B<blob_read>

  $blob = $sth->blob_read($id, $offset, $len);

Supported by this driver as proposed by DBI. Implemented by DBI 
but not documented, so this method might change. 

The offset and len parameter may be set to zero, in which case the driver 
fetches the whole blob at once. See also the B<blob_copy_to_file> method, 
which also supported by DBI, but not documented.

=back

=head2 Statement Handle Attributes

=over 4

=item B<NUM_OF_FIELDS>  (integer, read-only)

Implemented by DBI, no driver-specific impact.

=item B<NUM_OF_PARAMS>  (integer, read-only)

Implemented by DBI, no driver-specific impact. 

=item B<NAME>  (array-ref, read-only)

Supported by this driver as proposed by DBI.

=item B<NAME_lc>  (array-ref, read-only)

Implemented by DBI, no driver-specific impact.

=item B<NAME_uc>  (array-ref, read-only)

Implemented by DBI, no driver-specific impact.

=item B<NAME_hash>  (hash-ref, read-only)

Implemented by DBI, no driver-specific impact.

=item B<NAME_lc_hash>  (hash-ref, read-only)

Implemented by DBI, no driver-specific impact.

=item B<NAME_uc_hash>  (hash-ref, read-only)

Implemented by DBI, no driver-specific impact.

=item B<TYPE>  (array-ref, read-only)

Supported by this driver as proposed by DBI, with 
the restriction, that the types are FrontBase 
specific data-types which do not correspond to the ODBC
international standards.

=item B<PRECISION>  (array-ref, read-only)

Not supported by this driver. 

=item B<SCALE>  (array-ref, read-only)

Not supported by this driver. 

=item B<NULLABLE>  (array-ref, read-only)

Not supported by this driver. 

=item B<CursorName>  (string, read-only)

Not supported by this driver.

=item B<Database>  (dbh, read-only)

Implemented by DBI, no driver-specific impact.

=item B<ParamValues>  (hash ref, read-only)

Not supported by this driver.

=item B<ParamArrays>  (hash ref, read-only)

Not supported by this driver.

=item B<ParamTypes>  (hash ref, read-only)

Not supported by this driver.

=item B<Statement>  (string, read-only)

Supported by this driver as proposed by DBI. 

=item B<RowsInCache>  (integer, read-only)

Not supported by this driver.

=item B<fb_size>  (array-ref, read-only)

Frontbase specific attribute. It returns a reference to an 
array of integer values for each column in the result set. 
The integer shows the size of the column in bytes except for
BIT, VBIT and BYTE columns which are returned as the number of bits.
Note, that DATE, TIME, TIMESTAMP and their ZONEd equivalents are converted
to strings and the byte size is therefore the size of the result string.

=item B<fb_type>  (hash-ref, read-only)

Frontbase specific attribute. It returns a reference to an 
array of strings for each column. The string shows the name 
of the data_type. Note, that BYTE columns are described as BIT since 
BYTE(n) is equivalent to BIT(n*8).

=back


=head1 FURTHER INFORMATION

=head2 Transactions

Transaction behavior is controlled via the C<AutoCommit> attribute. For a
complete definition of C<AutoCommit> please refer to the DBI documentation.

According to the DBI specification the default for C<AutoCommit> is a true
value. In this mode, any change to the database becomes valid immediately. Any
C<BEGIN>, C<COMMIT> or C<ROLLBACK> statements will be rejected. The BEGIN 
statement is not used by FrontBase. If there is no active transaction then a 
new transaction is automatically started by the database when a statement is issued.

=head2 Type mappings

=head3 Boolean

The driver expects boolean values to be passed as unquoted string values. So true values are passed
as TRUE and false as FALSE. It does not currently support binding as SQL_BOOLEAN.

=head1 SEE ALSO

L<DBI>

=head1 AUTHORS

DBI by Tim Bunce (Tim.Bunce@ig.co.uk)

DBD-FB created by Anjo Krank (ak@prnet.de). By Colin Richardson <colin@frontbase.com> 
and others after 0.04.

This package was originally copied from DBD-Pg.

DBD-Pg by Edmund Mergl (E.Mergl@bawue.de) and Jeffrey W. Baker
(jwbaker@acm.org). By David Wheeler <david@justatheory.com>, Jason
Stewart <jason@openinformatics.com>, Bruce Momjian <pgman@candle.pha.pa.us>, 
Greg Sabino Mullane <greg@turnstep.com>, and others after v1.13.

B<Bug Reports>

If you feel certain you have found a bug, you can report it by sending
an email to <bugs@frontbase.com>.

=head1 COPYRIGHT

The DBD::FB module is free software. You may distribute under the terms of
either the GNU General Public License or the Artistic License, as specified in
the Perl README file, with the exception that it cannot be placed on a CD-ROM or 
similar media for commercial distribution without the prior approval of the author.

=head1 ACKNOWLEDGMENTS

B<Edmund Mergl (E.Mergl@bawue.de)> who has written the DBD::Pg Driver.
This driver wouldn't have been possible without his work.

B<Geert B. Clemmensen, (gclem@frontbase.com)> for creating Frontbase 
in the first place and his extraordinary support.

See also B<DBI/ACKNOWLEDGMENTS>.

=cut
