/*

 Copyright (c) 2000-2007 FrontBase, Inc.
 Portions Copyright (c) 1997,1998,1999 Edmund Mergl
 Portions Copyright (c) 1994,1995,1996,1997 Tim Bunce

 You may distribute under the terms of either the GNU General Public
 License or the Artistic License, as specified in the Perl README file.

 */

#include "FB.h"

/* another stupid workaround */
#ifndef YES
#define YES 1
#define NO 0
#endif

/* comment/uncomment the following to change the fetching behaviour*/
//#define USE_RH_ONCE 1
#undef USE_RH_ONCE
#define ROWS_PER_FETCH	(1000)	/* arbitrary number of rows to fetch per roundtrip */
typedef FBCMetaData PGresult;
typedef int ExecStatusType;

#define PGRES_COMMAND_OK 1
#define PGRES_TUPLES_OK 2
#define PGRES_UNIQUE_OK 3

#define FB_ERR_NO_CONNECTION 1
#define FB_BLOB_HANDLE_LENGTH 30
//cr137 start
#define FB_LONGINT_MAX_BYTES 8
#define FB_DATE_MAX_CHARS 10
#define FB_TIME_MAX_CHARS 8
#define FB_TIMESTAMP_MAX_CHARS 19
#define FB_TIME_WITH_ZONE_MAX_CHARS 14
#define FB_TIMESTAMP_WITH_ZONE_MAX_CHARS 25
#define FB_LOB_MAX_BYTES 2147483647
#define FB_BOOL_MAX_CHARS 5				// "FALSE"
#define PERL_BOOL_MAX_CHARS 1			// "0"/"1"
#define FB_INTEGER_MAX_DIGITS 10
#define FB_TINYINT_MAX_DIGITS 3
#define FB_SMALLINT_MAX_DIGITS 5
#define FB_LONGINT_MAX_DIGITS 19
#define FB_DOUBLE_MAX_DIGITS 15
//cr137 end

/* XXX DBI should provide a better version of this */
#define IS_DBI_HANDLE(h)  (SvROK(h) && SvTYPE(SvRV(h)) == SVt_PVHV && SvRMAGICAL(SvRV(h)) && (SvMAGIC(SvRV(h)))->mg_type == 'P')

//cr139
#define FB_SQLIdentifier 99

DBISTATE_DECLARE;

static SV *fb_pad_empty;

static void dbd_preparse(imp_sth_t *imp_sth, char *statement);
void upper_string(char *s);

//cr135 start
void		fssInitialize();
unsigned	fbaCount(FBArray *a);
void		*fbaObjectAtIndex(FBArray *a, unsigned indx);
void		fbaRelease(FBArray *a);
//cr135 end

#if WIN32
#define strncasecmp(x, y, z) strnicmp(x, y, z)
#endif

#ifdef free
#undef free
#endif

#ifdef malloc
#undef malloc
#endif

/* Database specific error handling.
This will be split up into specific routines
for dbh and sth level.
Also split into helper routine to set number & string.
Err, many changes needed, ramble ...
*/

void fb_error(SV *h, int error_num, const char *error_msg)
{
    D_imp_xxh(h);

    sv_setiv(DBIc_ERR(imp_xxh), (IV)error_num);		/* set err early */
    sv_setpv(DBIc_ERRSTR(imp_xxh), (char*)error_msg);
    DBIh_EVENT2(h, ERROR_event, DBIc_ERR(imp_xxh), DBIc_ERRSTR(imp_xxh));
    if (dbis->debug >= 2) {
        PerlIO_printf(DBILOGFP, "%s error %d recorded: %s\n", error_msg, error_num, SvPV(DBIc_ERRSTR(imp_xxh),na));
    }
}

static void dbd_drv_error(SV *h, int error_num, const char *error_msg)
{
	D_imp_xxh(h);
	DBIh_SET_ERR_CHAR(h, imp_xxh, Nullch, error_num, error_msg, Nullch, Nullch);
	if (DBIc_TRACE_LEVEL(imp_xxh) >= 2)
          PerlIO_printf(DBIc_LOGPIO(imp_xxh), "Error %s: %s\n", error_num, neatsvpv(newSVpv(error_msg,0),0));
}


static void dbd_drv_warn(SV *h, const char *warn_msg)
{
	D_imp_xxh(h);
	DBIh_SET_ERR_CHAR(h, imp_xxh, Nullch, 0, warn_msg, Nullch, Nullch);
	if (DBIc_TRACE_LEVEL(imp_xxh) >= 2)
          PerlIO_printf(DBIc_LOGPIO(imp_xxh), "Warning %s\n", neatsvpv(newSVpv(warn_msg,0),0));
}

//cr137 added
static int sql_type_from_fb_type(int fb)
{
	int retval = SQL_UNKNOWN_TYPE;

	switch(fb) {
		case FB_PrimaryKey:
			retval = SQL_INTEGER;
			break;
		case FB_Boolean:
			retval = SQL_BOOLEAN;
			break;
		case FB_Integer:
			retval = SQL_INTEGER;
			break;
		case FB_SmallInteger:
		case FB_TinyInteger:
			retval = SQL_SMALLINT;
			break;
		case FB_LongInteger:
			retval = SQL_BIGINT;
			break;
		case FB_Float:
		case FB_Real:
			retval = SQL_FLOAT;
			break;
		case FB_Double:
			retval = SQL_DOUBLE;
			break;
		case FB_Numeric:
			retval = SQL_NUMERIC;
			break;
		case FB_Decimal:
			retval = SQL_DECIMAL;
			break;
		case FB_Character:
			retval = SQL_CHAR;
			break;
		case FB_VCharacter:
			retval = SQL_VARCHAR;
			break;
		case FB_Bit:
			retval = SQL_BINARY;
			break;
		case FB_VBit:
			retval = SQL_VARBINARY;
			break;
		case FB_Date:
			retval = SQL_TYPE_DATE;
			break;
		case FB_Time:
		case FB_TimeTZ:
			retval = SQL_TYPE_TIME;
			break;
		case FB_Timestamp:
		case FB_TimestampTZ:
			retval = SQL_TYPE_TIMESTAMP;
			break;
		case FB_YearMonth:
			retval = SQL_INTERVAL_YEAR_TO_MONTH;
			break;
		case FB_DayTime:
			retval = SQL_INTERVAL_DAY_TO_SECOND;
			break;
		case FB_CLOB:
			retval = SQL_LONGVARCHAR;
			break;
		case FB_BLOB:
			retval = SQL_LONGVARBINARY;
			break;

		default:
			if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "sql_type_from_fb_type: Unknown server type %d\n", fb); }
			retval = SQL_UNKNOWN_TYPE;
			break;
	}

	return retval;
}

static Bool setConnectInfo(imp_dbh_t *imp_dbh)
{
	int status = False;
	if(imp_dbh->conn != NULL && fbcdcConnected(imp_dbh->conn)) {
		FBCMetaData* md = fbcdcExecuteDirectSQL(imp_dbh->conn, "extract database;");
		FBCPList* pl = fbcplParse(fbcmdMessage(md));
		FBCPList* plFBVersion;
		const char* fbVersion;
		// Database name
		FBCPList* plDbFile = (FBCPList*)fbcplValueForKey((const FBCPList*)pl, "DB FILE");
		const char* dbFile = fbcplString(plDbFile);
		int pos = strlen(dbFile)-1, start = 0, end = 0, len = 0;
		char pathSep = '/';
		while(pos>=0) {
			if(dbFile[pos] == '.')
				end = pos;
			if(pos < end && dbFile[pos] == pathSep) {
				start = pos+1;
				break;
			}
			pos--;
		}
		if(start < end) {
			len = end-start;
			if(len > 255) len = 255;
			strncpy(imp_dbh->dbname, dbFile+start,len);
			imp_dbh->dbname[len] = '\0';
		}
		
		// Server version
		plFBVersion = (FBCPList*)fbcplValueForKey((const FBCPList*)pl, "FB VERSION");
		fbVersion = fbcplString(plFBVersion);
		strncpy(imp_dbh->fb_server_version, fbVersion, 255);
		imp_dbh->fb_server_version[255] = '\0';
		fbcmdRelease(md);
		fbcplRelease(pl);
		status = True;
	}
	return status;
}

/* Postgres Wrapper stuff */

static FBCMetaData *PQexec(FBCDatabaseConnection *conn, char *string)
{
    FBCMetaData *self = NULL;
    int len =  strlen(string);
    char *newCmd = malloc(len+2);
    if(newCmd)
    {
        memcpy(newCmd, string, len);
        newCmd[len] = ';';
        newCmd[len+1] = '\0';

        if (dbis->debug >= 1) {
            PerlIO_printf(DBILOGFP, "PQexec: >>%s<<\n", newCmd);
        }
        
        self = fbcdcExecuteDirectSQL(conn, newCmd);
        free(newCmd);
    }
    
    return self;
}

//cr137
static unsigned FBerrorCode(FBCDatabaseConnection *conn, FBCMetaData *res)
{
	unsigned errCode = 0;
	if(res)
    {
        FBCErrorMetaData* emd = fbcdcErrorMetaData(conn, res);
		if(emd) {
			errCode = fbcemdErrorCodeAtIndex(emd, 0);
			fbcemdRelease(emd);
		}
    }
	return errCode;
}

static const char *PQerrorMessage(FBCDatabaseConnection *conn, FBCMetaData *res)
{
    static char buf[4096];
    buf[0] = 0;
	//cr137 : now caters for warnings
    if(res)
    {
        FBCErrorMetaData* emd = fbcdcErrorMetaData(conn, res);
		if(emd) {
			char* msgs = fbcemdAllErrorMessages(emd);
			strncpy(buf, msgs, sizeof(buf));

			fbcemdRelease(emd);
			free(msgs);
		}
    }
    
    return buf;
}

static int PQexecWithStatus(FBCDatabaseConnection *conn, char *cmd)
{
    FBCMetaData *meta;
    int result = PGRES_COMMAND_OK;

    meta = PQexec(conn, cmd);
    
    if (fbcmdErrorsFound(meta) == YES) {
        result = 0;
    }
    
    fbcmdRelease(meta);
    return result;
}

static char *PQfname(FBCMetaData *res, int index)
{
    static char buf[256];
    const FBCColumnMetaData *meta = fbcmdColumnMetaDataAtIndex(res, index);

    buf[0] = 0;

    if (meta) {
        strncpy(buf, fbccmdLabelName(meta), sizeof(buf));
        /*
         if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "PQfname: res 0x%08lx, meta 0x%08lx, name %s\n", res,meta, buf); }
         fbccmdRelease(meta);
         if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "PQfname: meta released\n", res); }
         */
    }
    /* YES! I know this is ugly, but I couldn't think of something better */
    return buf;
}

// cr137 : BIT length for BIT types. BYTE length for others. DateTimes are strings.
static int PQfsize(FBCMetaData *res, int index)
{
    int size = -1, type = -1;
    const FBCDatatypeMetaData *meta;
    meta = fbcmdDatatypeMetaDataAtIndex(res, index);

    if (meta) {
		type = fbcdmdDatatypeCode(meta);
		//cr137 start
		switch(type) {
		case FB_PrimaryKey:
		case FB_Integer:
		case FB_YearMonth:
		case FB_DayTime:
			size = sizeof(signed long);
			break;
		case FB_Boolean:
			size = sizeof(unsigned char);
			break;
		case FB_TinyInteger:
			size = sizeof(signed char);
			break;
		case FB_SmallInteger:
			size = sizeof(signed short);
			break;
		case FB_LongInteger:
			size = FB_LONGINT_MAX_BYTES;
			break;
		case FB_Float:
		case FB_Real:
		case FB_Double:
		case FB_Numeric:
		case FB_Decimal:
			size = sizeof(double);
			break;
		case FB_Bit:
		case FB_VBit:
		case FB_Character:
		case FB_VCharacter:
			size = fbcdmdLength(meta);
			break;
		case FB_Date:
			size = FB_DATE_MAX_CHARS;
			break;
		case FB_Time:
			size = FB_TIME_MAX_CHARS;
			break;
		case FB_TimeTZ:
			size = FB_TIME_WITH_ZONE_MAX_CHARS;
			break;
		case FB_Timestamp:
			size = FB_TIMESTAMP_MAX_CHARS;
			break;
		case FB_TimestampTZ:
			size = FB_TIMESTAMP_WITH_ZONE_MAX_CHARS;
			break;
		case FB_CLOB:
		case FB_BLOB:
			size = -1;
			break;
		default:
			if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "sql_type_from_fb_type: Unknown server type %d\n", type); }
			break;
		}
		//cr137 end
		if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "PQfsize: res 0x%08lx, meta 0x%08lx, size %d\n", res,meta, size); }
    }
    
    return size;
}

// cr137
static int FBcolPrec(imp_sth_t *imp_sth, int index)
{
	D_imp_dbh_from_sth;
    int size = -1, type = -1;
    const FBCDatatypeMetaData *meta;
    meta = fbcmdDatatypeMetaDataAtIndex(imp_sth->result, index);

    if (meta) {
		type = fbcdmdDatatypeCode(meta);
		//cr137 start
		switch(type) {
		case FB_PrimaryKey:
		case FB_Integer:
		case FB_YearMonth:
		case FB_DayTime:
			size = FB_INTEGER_MAX_DIGITS;
			break;
		case FB_Boolean:
			if(imp_dbh->fb_bools)
				size = FB_BOOL_MAX_CHARS;
			else
				size = PERL_BOOL_MAX_CHARS;
			break;
		case FB_TinyInteger:
			size = FB_TINYINT_MAX_DIGITS;
			break;
		case FB_SmallInteger:
			size = FB_SMALLINT_MAX_DIGITS;
			break;
		case FB_LongInteger:
			size = FB_LONGINT_MAX_DIGITS;
			break;
		case FB_Float:
		case FB_Real:
		case FB_Double:
			size = FB_DOUBLE_MAX_DIGITS;
			break;
		case FB_Numeric:
		case FB_Decimal:
			size = fbcdmdPrecision(meta);
			break;
		case FB_Bit:
		case FB_VBit:
		case FB_Character:
		case FB_VCharacter:
			size = fbcdmdLength(meta);
			break;
		case FB_Date:
			size = FB_DATE_MAX_CHARS;
			break;
		case FB_Time:
			size = FB_TIME_MAX_CHARS;
			break;
		case FB_TimeTZ:
			size = FB_TIME_WITH_ZONE_MAX_CHARS;
			break;
		case FB_Timestamp:
			size = FB_TIMESTAMP_MAX_CHARS;
			break;
		case FB_TimestampTZ:
			size = FB_TIMESTAMP_WITH_ZONE_MAX_CHARS;
			break;
		case FB_CLOB:
		case FB_BLOB:
			size = FB_LOB_MAX_BYTES;
			break;
		default:
			if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "FBcolPrec: Unknown server type %d\n", type); }
			break;
		}
		//cr137 end
		if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "FBcolPrec: res 0x%08lx, meta 0x%08lx, size %d\n", imp_sth->result ,meta, size); }
    }
    
    return size;
}

// cr137
static int FBcolScale(FBCMetaData* res, int index)
{
    int scale = -1, type = -1;
    const FBCDatatypeMetaData *meta;
    meta = fbcmdDatatypeMetaDataAtIndex(res, index);

    if (meta) {
		type = fbcdmdDatatypeCode(meta);
		//cr137 start
		switch(type) {
		case FB_Numeric:
		case FB_Decimal:
			scale = fbcdmdScale(meta);
			break;
		case FB_PrimaryKey:
		case FB_Integer:
		case FB_YearMonth:
		case FB_DayTime:
		case FB_Boolean:
		case FB_TinyInteger:
		case FB_SmallInteger:
		case FB_LongInteger:
		case FB_Float:
		case FB_Real:
		case FB_Double:
		case FB_Bit:
		case FB_VBit:
		case FB_Character:
		case FB_VCharacter:
		case FB_Date:
		case FB_Time:
		case FB_TimeTZ:
		case FB_Timestamp:
		case FB_TimestampTZ:
		case FB_CLOB:
		case FB_BLOB:
		default:
			if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "FBcolScale: Unknown server type %d\n", type); }
			break;
		}
		//cr137 end
		if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "FBcolScale: res 0x%08lx, meta 0x%08lx, scale %d\n", res ,meta, scale); }
    }
    
    return scale;
}

static int PQftype(FBCMetaData *res, int index)
{
    int type = -1;
    const FBCDatatypeMetaData *meta = fbcmdDatatypeMetaDataAtIndex(res, index);

    if (meta) {
        type = fbcdmdDatatypeCode(meta);
		if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "PQftype: res 0x%08lx, meta 0x%08lx, type %d\n", res,meta, type); }
    }
    
    return type;
}

static int convert_to_boolean(char* value, char* buffer)
{
	if(strcmp(value, "0") == 0 || strcmp(value, "") == 0 || strcmp(value, "FALSE") == 0)
		strcpy(buffer, "FALSE");
	else
		strcpy(buffer, "TRUE");
	return 1;
}

void dbd_init(dbistate_t *dbistate)
{
    DBIS = dbistate;

    fssInitialize();
    setlinebuf(stdout);

    if (getenv("FB_PAD_EMPTY")) {
        sv_setiv(fb_pad_empty, atoi(getenv("FB_PAD_EMPTY")));
    }
}


int dbd_discon_all(SV *drh, imp_drh_t *imp_drh)
{
    dTHR;

    if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "dbd_discon_all\n"); }

    /* The disconnect_all concept is flawed and needs more work */
    if (!dirty && !SvTRUE(perl_get_sv("DBI::PERL_ENDING",0))) {
        sv_setiv(DBIc_ERR(imp_drh), (IV)1);
        sv_setpv(DBIc_ERRSTR(imp_drh), (char*)"disconnect_all not implemented");
        DBIh_EVENT2(drh, ERROR_event, DBIc_ERR(imp_drh), DBIc_ERRSTR(imp_drh));
        return FALSE;
    }
    if (perl_destruct_level) {
        perl_destruct_level = 0;
    }
    return FALSE;
}


static int fbtype_bind_ok(int dbtype)	/* It's a type we support for placeholders */
{
	//crLOB, cr137 : extended to include Tiny and LongInts
	//c139 : added FB_SQLIdentifier for binding unquoted varchar identifiers
    return (dbtype >= FB_Boolean && dbtype <= FB_LongInteger || dbtype == FB_SQLIdentifier ?  1: 0); 
}

int dbd_db_login(SV *dbh, imp_dbh_t *imp_dbh, char *dbname, char *uid, char *pwd)
{
    dTHR;

    char *conn_str;
    char *src;
    char *dbpasswd;
   	FBCMetaData* meta;
    char *host;
	Bool usePort = True;	//cr137 start
	long portNumber = 0;
	int i = 0;
	char *upper, digested[17];		//cr137 end
	
    if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "dbd_db_login >%s<\n", dbname); }

    /* build connect string */
    /* DBD-FB syntax: 'dbname=dbname;host=host' */
    /* fbsql  syntax: 'dbname/host/dbpasswd' */

    conn_str = (char *)malloc(strlen(dbname)+1);
    if (! conn_str) {
        return 0;
    }
    strcpy(conn_str, dbname);

    src  = conn_str;
    while (*src) {
        if (*src != '/') {
            src++;
            continue;
        }
        *src++ = '\0';
        host = src;
        break;
    }
    dbname = conn_str;
    while (*src) {
        if (*src != '/') {
            src++;
            continue;
        }
        *src++ = '\0';
        dbpasswd = src;
        break;
    }
	
	//cr137 start
	while(dbname[i] != '\0') {
		if(!isdigit(dbname[i++])) {
			usePort = False;
			break;
		}
	}

	if(usePort) {
		portNumber = atoi(dbname);
		if(!(portNumber > 0 && portNumber <=65535))
			usePort = False;
	}
	//cr137 end
    if (dbis->debug >= 2) { PerlIO_printf(DBILOGFP, "dbd_db_login: conn to db >%s< at host >%s< with dbpasswd >%s<\n", dbname, host, dbpasswd); }


    /* make a connection to the database */
    if(dbname && *dbname)
    {
		//cr137 start
		if(!dbpasswd || *dbpasswd == '\0')
			digested[0] = '\0';
		else
			fbcDigestPassword("_SYSTEM", dbpasswd, digested);
		
		if(usePort)
			imp_dbh->conn = fbcdcConnectToDatabaseUsingPort(host, portNumber, digested);
		else
			imp_dbh->conn = fbcdcConnectToDatabase(dbname, host, digested);
		//cr137 end
        free(conn_str);
        if (imp_dbh->conn == NULL) {
            dbd_drv_error(dbh, FB_ERR_NO_CONNECTION, fbcdcClassErrorMessage());
            return 0;
        }
        /* dirk: set the charater set for this connection */
        //fbcdcSetInputCharacterSet(imp_dbh->conn, FBC_ISO8859_1); //cr134 set UTF8 as default
		// UTF8 is the expected input and output character encoding
		// Comment these lines out to use latin1 on Win32.
		fbcdcSetInputCharacterSet(imp_dbh->conn, FBC_Undefined);
		fbcdcSetOutputCharacterSet(imp_dbh->conn, FBC_Undefined);
		
		if(pwd && *pwd != '\0') {
			upper = strdup(uid);
			upper_string(upper); 
			fbcDigestPassword( (const char *)upper, (const char *)pwd, digested);
			free(upper);
		} 
		else {
			digested[0] = '\0';
		}
	
        meta = fbcdcCreateSession(imp_dbh->conn, "Perl::DBI", uid, digested, "system_user");

        if (fbcmdErrorsFound(meta) == YES) {
            FBCErrorMetaData* emd = fbcdcErrorMetaData(imp_dbh->conn, meta);
            char* msgs = fbcemdAllErrorMessages(emd);

            dbd_drv_error(dbh, FB_ERR_NO_CONNECTION, msgs);
            fbcemdRelease(emd);
            free(msgs);

            fbcmdRelease(meta);

            fbcdcClose(imp_dbh->conn);
            fbcdcRelease(imp_dbh->conn);
            imp_dbh->conn = NULL;

            return 0;
        }
        fbcmdRelease(meta);
    }
    else
    {
        imp_dbh->execHandler = fbcehHandlerForHost(host, FBCDatabaseManager);
        if (imp_dbh->execHandler == NULL) {
            dbd_drv_error(dbh, FB_ERR_NO_CONNECTION, fbcehClassErrorMessage());
            return 0;
        }
    }

    imp_dbh->init_commit = 1;			/* initialize AutoCommit */
    imp_dbh->fb_auto_escape = 1;		/* initialize fb_auto_escape */
	imp_dbh->fb_bools = 1;				/* initialize fb_bools */
	//cr137 start
	strncpy(imp_dbh->host, host, 255);
	imp_dbh->host[255] = '\0';
	strncpy(imp_dbh->fb_username, uid, 255);
	imp_dbh->fb_username[255] = '\0';
	setConnectInfo(imp_dbh);			
	//cr137 end
    DBIc_IMPSET_on(imp_dbh);			/* imp_dbh set up now */
    DBIc_ACTIVE_on(imp_dbh);			/* call disconnect before freeing */
    return 1;
}

//cr136 switched behaviour as this makes more sense and fixes connect_cached
int dbd_db_ping(SV *dbh)
{
    D_imp_dbh(dbh);

    if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "dbd_db_ping\n"); }

    if (fbcdcConnected(imp_dbh->conn) == YES) {
        return 1;
    }

    return 0;
}


int dbd_db_commit(SV *dbh, imp_dbh_t *imp_dbh)
{
    PGresult* result = 0;
    ExecStatusType status;
    int retval = 1;

    if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "dbd_db_commit\n"); }

    /* no commit if AutoCommit = on */
    if (DBIc_has(imp_dbh, DBIcf_AutoCommit) != FALSE) {
        return 0;
    }

    /* execute commit */
    status = PQexecWithStatus(imp_dbh->conn, "commit");

    if (status != PGRES_COMMAND_OK) {
        dbd_drv_error(dbh, status, "commit failed\n");
        return 0;
    }

    if (DBIc_has(imp_dbh, DBIcf_AutoCommit) == FALSE) {
    	/*
    	//	"begin" does not begin a transaction. It begins a statement block.
    	//	so this was causing transactions to blow up. Nice. -Brad Hutchings 03-Oct-2001
    	
        status = PQexecWithStatus(imp_dbh->conn, "begin");
        if (status != PGRES_COMMAND_OK  && status != 0) {
            dbd_drv_error(dbh, status, "begin failed\n");
            return 0;
        }
        */
    }

    return retval;
}


int dbd_db_rollback(SV *dbh, imp_dbh_t *imp_dbh)
{
    PGresult* result = 0;
    ExecStatusType status;
    int retval = 1;

    if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "dbd_db_rollback\n"); }

    /* no rollback if AutoCommit = on */
    if (DBIc_has(imp_dbh, DBIcf_AutoCommit) != FALSE) {
        return 0;
    }

    /* execute rollback */
    status = PQexecWithStatus(imp_dbh->conn, "rollback");

    if (status != PGRES_COMMAND_OK) {
        dbd_drv_error(dbh, status, "rollback failed\n");
        return 0;
    }

    if (DBIc_has(imp_dbh, DBIcf_AutoCommit) == FALSE) {
    	/*
    	//	"begin" does not begin a transaction. It begins a statement block.
    	//	so this was causing transactions to blow up. Nice. -Brad Hutchings 03-Oct-2001
        status = PQexecWithStatus(imp_dbh->conn, "begin");
        if (status != PGRES_COMMAND_OK  && status != 0) {
            dbd_drv_error(dbh, status, "begin failed\n");
            return 0;
        }
        */
    }

    return retval;
}


int dbd_db_disconnect(SV *dbh, imp_dbh_t *imp_dbh)
{
    dTHR;
    ExecStatusType status;

    if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "dbd_db_disconnect\n"); }

    /* We assume that disconnect will always work	*/
    /* since most errors imply already disconnected.	*/
    DBIc_ACTIVE_off(imp_dbh);

    /* rollback if AutoCommit = off */
	// think server will do this automatically (cr).
    if (DBIc_has(imp_dbh, DBIcf_AutoCommit) == FALSE) {
        status = PQexecWithStatus(imp_dbh->conn, "rollback");
        if (status != PGRES_COMMAND_OK) {
            dbd_drv_error(dbh, status, "rollback failed\n");
            return 0;
        }
        if (dbis->debug >= 2) { PerlIO_printf(DBILOGFP, "dbd_db_disconnect: AutoCommit=off -> rollback\n"); }
    }

    if(imp_dbh->conn)
    {
        fbcdcClose(imp_dbh->conn);
    }
    if(imp_dbh->execHandler)
    {
        fbcehRelease(imp_dbh->execHandler);
        imp_dbh->execHandler = 0;

    }
    /* We don't free imp_dbh since a reference still exists	*/
    /* The DESTROY method is the only one to 'free' memory.	*/
    /* Note that statement objects may still exists for this dbh!	*/
    return 1;
}


void dbd_db_destroy(SV *dbh, imp_dbh_t *imp_dbh)
{
    if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "dbd_db_destroy\n"); }

    if (DBIc_ACTIVE(imp_dbh)) {
        dbd_db_disconnect(dbh, imp_dbh);
    }

    /* Nothing in imp_dbh to be freed	*/
    DBIc_IMPSET_off(imp_dbh);
}


int dbd_db_STORE_attrib(SV *dbh, imp_dbh_t *imp_dbh, SV *keysv, SV *valuesv)
{
    STRLEN kl;
    char *key = SvPV(keysv,kl);
    int newval = SvTRUE(valuesv);
    ExecStatusType status;

    if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "dbd_db_STORE\n"); }

    if (kl==10 && strEQ(key, "AutoCommit")) {
        int oldval = DBIc_has(imp_dbh, DBIcf_AutoCommit);
        DBIc_set(imp_dbh, DBIcf_AutoCommit, newval);
        if(newval != oldval && DBIc_ACTIVE(imp_dbh)) {
			// AUTOCOMMIT ON
			if(newval) {
				if (dbis->debug >= 2) { PerlIO_printf(DBILOGFP, "dbd_db_STORE: switch AutoCommit to on\n"); }
				status = PQexecWithStatus(imp_dbh->conn, "COMMIT;SET COMMIT TRUE");
				if (status != PGRES_COMMAND_OK) {
					dbd_drv_error(dbh, status, "COMMIT;SET COMMIT TRUE failed\n");
					return 0;
				}
			}
			// AUTOCOMMIT OFF
			else {
				if (dbis->debug >= 2) { PerlIO_printf(DBILOGFP, "dbd_db_STORE: switch AutoCommit to off\n"); }
				status = PQexecWithStatus(imp_dbh->conn, "SET COMMIT FALSE");
				if (status != PGRES_COMMAND_OK) {
					dbd_drv_error(dbh, status, "SET COMMIT FALSE failed\n");
					return 0;
				}
			}
		}
		return 1;
    } else if (kl==14 && strEQ(key, "fb_auto_escape")) {
        imp_dbh->fb_auto_escape = newval;
		return 1; //cr0135
    } else if(kl==12 && strEQ(key, "fb_bools")) {	//cr137 added for backwards compatibility
		imp_dbh->fb_bools = newval;
		return 1;
	}
	return 0;
}


SV *dbd_db_FETCH_attrib(SV *dbh, imp_dbh_t *imp_dbh, SV *keysv)
{
    STRLEN kl;
    char *key = SvPV(keysv,kl);
    SV *retsv = Nullsv;

    if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "dbd_db_FETCH\n"); }

    if (kl==10 && strEQ(key, "AutoCommit")) {
        retsv = boolSV(DBIc_has(imp_dbh, DBIcf_AutoCommit));
	} else if(kl==4 && strEQ(key, "Name")) {	//cr137
		 retsv = newSVpv(imp_dbh->dbname, 0);
    } else if (kl==14 && strEQ(key, "fb_auto_escape")) {
        retsv = newSViv((IV)imp_dbh->fb_auto_escape);
    } else if (kl==15 && strEQ(key, "fb_data_sources")) {
        AV *av = newAV();
        int cnt, i;

        FBArray *arr;
        arr = fbcehAvailableDatabases(imp_dbh->execHandler);
		cnt = fbaCount(arr);
        retsv = newRV(sv_2mortal((SV*)av));
        for (i = 0; i <= cnt; i++) {
            char *type_nam = (char *)fbaObjectAtIndex(arr, i);
            if(!type_nam)
                continue;

            if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "fb_data_sources %s\n", type_nam); }
            av_store(av, i, newSVpv(type_nam, 0));
        }
        fbaRelease(arr);
	//cr137 start
    } else if(kl==12 && strEQ(key, "fb_bools")) {	
		 retsv = newSViv((IV)imp_dbh->fb_bools);
	} else if(kl==17 && strEQ(key, "fb_server_version")) {
		retsv = newSVpv(imp_dbh->fb_server_version, 0);
	} else if(kl==11 && strEQ(key, "fb_username")) {
		retsv = newSVpv(imp_dbh->fb_username, 0);
	}

	//cr137 end
    if (retsv == &sv_yes || retsv == &sv_no) {
        return retsv; /* no need to mortalize yes or no */
    }
    
    return sv_2mortal(retsv);
}


/* ================================================================== */


int dbd_st_prepare(SV *sth, imp_sth_t *imp_sth, char *statement, SV *attribs)
{
    if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "dbd_st_prepare: statement = >%s<\n", statement); }

    /* scan statement for '?', ':1' and/or ':foo' style placeholders */
    dbd_preparse(imp_sth, statement);

    /* initialize new statement handle */
    imp_sth->result    = 0;
    imp_sth->current_row = 0;
	imp_sth->rows = -1; //cr137 was defaulting to 0

    DBIc_IMPSET_on(imp_sth);
    return 1;
}


static void dbd_preparse(imp_sth_t *imp_sth, char *statement)
{
    bool in_literal = FALSE;
    char *src, *start, *dest;
    phs_t phs_tpl;
    SV *phs_sv;
    int idx=0;
    char *style="", *laststyle=Nullch;
    STRLEN namelen;

    if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "dbd_st_preparse: statement = >%s<\n", statement); }

    /* allocate room for copy of statement with spare capacity	*/
    /* for editing '?' or ':1' into ':p1'.			*/
    imp_sth->statement = (char*)safemalloc(strlen(statement) * 3);

    /* initialise phs ready to be cloned per placeholder	*/
    memset(&phs_tpl, 0, sizeof(phs_tpl));
    phs_tpl.ftype = FB_VCharacter;

    src  = statement;
    dest = imp_sth->statement;
    while(*src) {
        if (*src == '\'') {
            in_literal = ~in_literal;
        }
        
        // keep quotes
        if (in_literal && *src == '\\') {
            *dest++ = *src++;
        }
        
        // check for placeholders but take care of cast operator
        if ((*src != ':' && *src != '?') || (*src == ':' && (*(src-1) == ':' || *(src+1) == ':')) || in_literal) {
            *dest++ = *src++;
            continue;
        }

        // save name inc colon
        start = dest;
        *dest++ = *src++;
        
        if (*start == '?') {
            // X/Open standard
            sprintf(start,":p%d", ++idx); // '?' -> ':p1' (etc)
            dest = start+strlen(start);
            style = "?";
        } else if (isDIGIT(*src)) { // ':1'
            idx = atoi(src);
            *dest++ = 'p';		// ':1'->':p1'
            if (idx <= 0) {
                croak("Placeholder :%d invalid, placeholders must be >= 1", idx);
            }
            while(isDIGIT(*src)) {
                *dest++ = *src++;
            }
            style = ":1";
        } else if (isALNUM(*src)) {	// ':foo'
            while(isALNUM(*src)) {	// includes '_'
                *dest++ = *src++;
            }
            style = ":foo";
        } else { // perhaps ':=' PL/SQL construct
            continue;
        }
        *dest = '\0'; // handy for debugging
        namelen = (dest-start);
        if (laststyle && style != laststyle) {
            croak("Can't mix placeholder styles (%s/%s)",style,laststyle);
        }
        laststyle = style;
        if (imp_sth->all_params_hv == NULL) {
            imp_sth->all_params_hv = newHV();
        }
        phs_tpl.sv = &sv_undef;
        phs_sv = newSVpv((char*)&phs_tpl, sizeof(phs_tpl)+namelen+1);
        hv_store(imp_sth->all_params_hv, start, namelen, phs_sv, 0);
        strcpy( ((phs_t*)(void*)SvPVX(phs_sv))->name, start);
    }
    *dest = '\0';
    if (imp_sth->all_params_hv) {
        DBIc_NUM_PARAMS(imp_sth) = (int)HvKEYS(imp_sth->all_params_hv);
        if (dbis->debug >= 2) { PerlIO_printf(DBILOGFP, "    dbd_preparse scanned %d distinct placeholders\n", (int)DBIc_NUM_PARAMS(imp_sth)); }
    }
}

//cr137 updated to match correct FB_xxx types
static int fb_sql_type(imp_sth_t *imp_sth, char *name, int sql_type)
{
    switch (sql_type) {
        case SQL_NUMERIC:
			return FB_Numeric;
        case SQL_DECIMAL:
			return FB_Decimal;
        case SQL_INTEGER:
			return FB_Integer;
        case SQL_SMALLINT:
			return FB_SmallInteger;
        case SQL_FLOAT:
			return FB_Float;
        case SQL_REAL:
			return FB_Real;
        case SQL_DOUBLE:
			return FB_Double;
        case SQL_VARCHAR:
            return FB_VCharacter;	
        case SQL_CHAR:
            return FB_Character;	
		//crLOB start
		case SQL_LONGVARBINARY:
			return FB_BLOB;
		case SQL_LONGVARCHAR:
			return FB_CLOB;
		//crLOB end
		//crDATETIME start
		case SQL_DATE:
		case SQL_TYPE_DATE:
			return FB_Date;
		case SQL_TIME:
		case SQL_TYPE_TIME:
			return FB_Time;
		case SQL_TIMESTAMP:
		case SQL_TYPE_TIMESTAMP:
			return FB_Timestamp;
		//crDATETIME end
		case SQL_BOOLEAN:			//cr137
			return FB_Boolean;
        default:
            if (imp_sth && DBIc_WARN(imp_sth) && name) {
                warn("SQL type %d for '%s' is not fully supported, bound as VARCHAR instead", sql_type, name); //cr0001 add args to stop crash
			}
            return fb_sql_type(imp_sth, name, SQL_VARCHAR);
    }
}


static int _dbd_rebind_ph(SV *sth, imp_sth_t *imp_sth, phs_t *phs)
{
    STRLEN value_len;

    if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "dbd_st_rebind\n"); }

    if (dbis->debug >= 2) {
        char *text;
        if(!SvOK(phs->sv)) {
            /* undef -> NULL */
            text = "NULL";
            PerlIO_printf(DBILOGFP, "bind %s <== %s\n",
                    phs->name, text);
        } else {
            text = neatsvpv(phs->sv, 0);
            PerlIO_printf(DBILOGFP, "bind %s <== %s (size %ld/%ld/%ld, ptype %ld, otype %d%s)\n",
                    phs->name, text, SvCUR(phs->sv),SvLEN(phs->sv),phs->maxlen,
                    SvTYPE(phs->sv), phs->ftype, (phs->is_inout) ? ", inout" : "");
        }
    }

    // At the moment we always do sv_setsv() and rebind.  Later we may optimise this so that more often we can just copy the value & length over and not rebind.

    if (phs->is_inout) {
        if (SvREADONLY(phs->sv)) {
            croak(no_modify);
        }
        if (imp_sth->fb_pad_empty) {
            croak("Can't use fb_pad_empty with bind_param_inout");
        }
        // phs->sv _is_ the real live variable, it may 'mutate' later pre-ufbrade high to reduce risk of SvPVX realloc/move
        SvUPGRADE(phs->sv, SVt_PVNV);
        // ensure room for result, 28 is magic number (see sv_2pv)
        SvGROW(phs->sv, (STRLEN)((phs->maxlen < 28) ? 28 : phs->maxlen+1));	//cr0135
    } else {
        // phs->sv is copy of real variable, upgrade to at least string
        SvUPGRADE(phs->sv, SVt_PV);
    }

    // At this point phs->sv must be at least a PV with a valid buffer, even if it's undef (null).  Here we set phs->progv, phs->indp, and value_len.
    if (SvOK(phs->sv)) {
        phs->progv = SvPV(phs->sv, value_len);
        phs->indp  = 0;
    } else {
        // it's null but point to buffer in case it's an out var
        phs->progv = SvPVX(phs->sv);
        phs->indp  = -1;
        value_len  = 0;
    }

    if (imp_sth->fb_pad_empty && value_len==0) {
        sv_setpv(phs->sv, " ");
        phs->progv = SvPV(phs->sv, value_len);
    }
    phs->sv_type = SvTYPE(phs->sv); // part of mutation check
    phs->maxlen  = SvLEN(phs->sv)-1; // avail buffer space

    phs->alen = value_len + phs->alen_incnull;

	switch(phs->ftype) {
		//crLOB start
		case FB_CLOB:
		case FB_BLOB:
			imp_sth->all_params_len += FB_BLOB_HANDLE_LENGTH;
			break;
		//crLOB end
		//crDATETIME start: add prefix lengths
		case FB_Date:
		case FB_Time:
			imp_sth->all_params_len += phs->alen+5;
			break;
		case FB_Timestamp:
			imp_sth->all_params_len += phs->alen+10;
			break;
		//crDATETIME end
		case FB_Boolean:
			imp_sth->all_params_len += 5; // cr137 make space for "FALSE"
			break;
		default:
			imp_sth->all_params_len += phs->alen;
	}
    if (dbis->debug >= 3) {
        PerlIO_printf(DBILOGFP, "bind %s <== '%.*s' (size %d/%ld, otype %d, indp %d)\n", phs->name, phs->alen, (phs->progv) ? phs->progv : "", phs->alen, (long)phs->maxlen, phs->ftype, phs->indp);
    }

    return 1;
}
 
int dbd_bind_ph(SV *sth, imp_sth_t *imp_sth, SV *ph_namesv, SV *newvalue, IV sql_type, SV *attribs, int is_inout, IV maxlen)
{
    SV **phs_svp;
    STRLEN name_len;
    char *name;
    char namebuf[30];
    phs_t *phs;

    if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "dbd_bind_ph\n"); }

    /* check if placeholder was passed as a number        */

    if (!SvNIOK(ph_namesv) && !SvPOK(ph_namesv)) {
        SvPV(ph_namesv, na); //force SvPOK
    }
    if (SvNIOK(ph_namesv) || (SvPOK(ph_namesv) && isDIGIT(*SvPVX(ph_namesv)))) {
        sprintf(namebuf, ":p%d", (int)SvIV(ph_namesv));
        name = namebuf;
        name_len = strlen(name);
    } else {
        // use the supplied placeholder name directly
        name = SvPV(ph_namesv, name_len);
        /* could check for leading colon here */
    }

    if (SvTYPE(newvalue) > SVt_PVMG) { // hook for later array logic
        croak("Can't bind non-scalar value (currently)");
    }
    if (SvROK(newvalue) && !IS_DBI_HANDLE(newvalue)) {	// dbi handle allowed for cursor variables
        croak("Can't bind a reference (%s)", neatsvpv(newvalue,0));
    }
    if (SvTYPE(newvalue) == SVt_PVLV && is_inout) {	// may allow later
        croak("Can't bind ``lvalue'' mode scalar as inout parameter (currently)");
    }

    if (dbis->debug >= 2) {
        PerlIO_printf(DBILOGFP, "         bind %s <== %s (type %ld", name, neatsvpv(newvalue,0), (long)sql_type);
        if (is_inout) {
            PerlIO_printf(DBILOGFP, ", inout 0x%lx", (long)newvalue);
        }
        if (attribs) {
            PerlIO_printf(DBILOGFP, ", attribs: %s", SvPV(attribs,na));
        }
        PerlIO_printf(DBILOGFP, ")\n");
    }

    phs_svp = hv_fetch(imp_sth->all_params_hv, name, name_len, 0);
    if (phs_svp == NULL) {
        croak("Can't bind unknown placeholder '%s'", name);
    }
    phs = (phs_t*)(void*)SvPVX(*phs_svp);	// placeholder struct

    if (phs->sv == &sv_undef) { // first bind for this placeholder
        phs->ftype    = FB_VCharacter;		 // our default type VARCHAR
        phs->maxlen   = maxlen;		 // 0 if not inout
        phs->is_inout = is_inout;
        if (is_inout) {
            // phs->sv assigned in the code below
            ++imp_sth->has_inout_params;
            // build array of phs's so we can deal with out vars fast
            if (!imp_sth->out_params_av)
                imp_sth->out_params_av = newAV();
            av_push(imp_sth->out_params_av, SvREFCNT_inc(*phs_svp));
        }

        if (attribs) {	// only look for fb_type on first bind of var
            SV **svp;
            // Setup / Clear attributes as defined by attribs.
            // XXX If attribs is EMPTY then reset attribs to default?
            if ( (svp = hv_fetch((HV*)SvRV(attribs), "fb_type", 7,  0)) != NULL) {
                int fb_type = SvIV(*svp);
                if (!fbtype_bind_ok(fb_type)) {
                    croak("Can't bind %s, fb_type %d not supported by DBD::FB", phs->name, fb_type);
                }
                if (sql_type) {
                    croak("Can't specify both TYPE (%d) and fb_type (%d) for %s", sql_type, fb_type, phs->name);
                }
                phs->ftype = fb_type;
            }
        }
        if (sql_type) {
            phs->ftype = fb_sql_type(imp_sth, phs->name, sql_type);
        }
    } else if (is_inout || phs->is_inout) {
        // check later rebinds for any changes
        croak("Can't rebind or change param %s in/out mode after first bind (%d => %d)", phs->name, phs->is_inout , is_inout);
    } else if (sql_type && phs->ftype != fb_sql_type(imp_sth, phs->name, sql_type)) {
        croak("Can't change TYPE of param %s to %d after initial bind", phs->name, sql_type);
    }

    if (!is_inout) {	// normal bind to take a (new) copy of current value
        if (phs->sv == &sv_undef)      // (first time bind)
            phs->sv = newSV(0);
        sv_setsv(phs->sv, newvalue);
    } else if (newvalue != phs->sv) {
        if (phs->sv) {
            SvREFCNT_dec(phs->sv);
        }
        phs->sv = SvREFCNT_inc(newvalue);	/* point to live var	*/
    }

    return _dbd_rebind_ph(sth, imp_sth, phs);
}

// <= -2:error, >=0:ok row count, (-1=unknown count)
//cr136 This needs to be reviewed. Maybe we should convert the value to the preferred type where that conversion can be supported, e.g. binary to hex string for FB_Bit
int dbd_st_execute(SV *sth, imp_sth_t *imp_sth)
{
    dTHR;

    D_imp_dbh_from_sth;
    char *statement;
    int ret = -2;
    int num_fields;
    int i;
    STRLEN len;
    bool in_literal = FALSE;
    char *src;
    char *dest;
    char *val;
    char namebuf[30];
    phs_t *phs;
    SV **svp;
	FBCBlobHandle* bh = NULL; //crLOB
	char buffer[1024]; //cr137 buffer for conversion 

    if (dbis->debug >= 1) {
        PerlIO_printf(DBILOGFP, "dbd_st_execute\n");
    }
	
	if (DBIc_ACTIVE(imp_sth))
      dbd_st_finish(sth, imp_sth);

    statement = imp_sth->statement;
    if (! statement) {
        // are we prepared ?
        dbd_drv_error(sth, -1, "statement not prepared\n");
        return -2;
    }

    // do we have input parameters?
    if ((int)DBIc_NUM_PARAMS(imp_sth) > 0) {
        // we have to allocate some additional memory for possible escaping quotes and backslashes
        statement = (char*)safemalloc(strlen(imp_sth->statement) + 2 * imp_sth->all_params_len);
        dest = statement;
        src  = imp_sth->statement;
        // scan statement for ':p1' style placeholders
        while(*src) {
            if (*src == '\'') {
                in_literal = ~in_literal;
            }
            if(in_literal && *src == '\\') {
                *dest++ = *src++;
            }
            if (*src != ':' || in_literal) {
                *dest++ = *src++;
                continue;
            }
            i = 0;
            namebuf[i++] = *src++; // ':'
            namebuf[i++] = *src++; // 'p'
            while (isDIGIT(*src)) {
                namebuf[i++] = *src++;
            }
            namebuf[i] = '\0';
            svp = hv_fetch(imp_sth->all_params_hv, namebuf, i, 0);
            if (svp == NULL) {
                dbd_drv_error(sth, -1, "parameter unknown\n");
                return -2;
            }
            // get attribute
            phs = (phs_t*)(void*)SvPVX(*svp);
            // replace undef with NULL
            if(!SvOK(phs->sv)) {
                val = "NULL";
                len = 4;
            } else {
                val = SvPV(phs->sv,len);
				switch(phs->ftype) {
					//crLOB start
					case FB_CLOB: {
						bh = fbcdcWriteCLOBWithLength(imp_dbh->conn, val, len);
						if(bh != NULL) {
							val = fbcbhDescription(bh);
							len = strlen(val);
						}
						else {
							val = "NULL";
							len = 4;
						}
						break;
					}
					case FB_BLOB: {
						bh = fbcdcWriteBLOB(imp_dbh->conn, val, len);
						if(bh != NULL) {
							val = fbcbhDescription(bh);
							len = strlen(val);
						}
						else {
							val = "NULL";
							len = 4;
						}
						break;
					}
					//crLOB end
					//crDATETIME start
					case FB_Date: {
						if(SvOK(phs->sv) && strncasecmp(val, "DATE", 4) != 0) {
							sprintf(buffer, "DATE '%s'", val);
							len = strlen(buffer);
							val = buffer;
						}
						break;
					}
					case FB_Time: {
						if(SvOK(phs->sv) && strncasecmp(val, "TIME", 4) != 0) {
							sprintf(buffer, "TIME '%s'", val);
							len = strlen(buffer);
							val = buffer;
						}
						break;
					}
					case FB_Timestamp: {
						if(SvOK(phs->sv) && strncasecmp(val, "TIMESTAMP", 9) != 0) {
							sprintf(buffer, "TIMESTAMP '%s'", val);
							len = strlen(buffer);
							val = buffer;
						}
						break;
					}
					case FB_Boolean : {
						convert_to_boolean(val, buffer);
						val = buffer;
						len = strlen(buffer);
						break;
					}
					default: {
					}
				}
			}
			// quote string attribute
			if(!SvNIOK(phs->sv) && SvOK(phs->sv) && (phs->ftype == FB_Character || phs->ftype == FB_VCharacter)) {	// avoid quoting NULL or LOB handles
					*dest++ = '\'';
			}
			//crDATETIME end
            while (len--) {
                if (imp_dbh->fb_auto_escape && (phs->ftype == FB_Character || phs->ftype == FB_VCharacter)) {
                    // escape quote
                    if (*val == '\'') {
                        *dest++ = '\'';
                    }
                    // escape backslash except for octal presentation
					// redundant //cr137
                    if (*val == '\\' && !isdigit(*(val+1)) && !isdigit(*(val+2)) && !isdigit(*(val+3)) ) {
                        *dest++ = '\\';
                    }
                }
                // copy attribute to statement
                *dest++ = *val++;
            }
            // quote string attribute
            if(!SvNIOK(phs->sv) && SvOK(phs->sv) && (phs->ftype == FB_Character || phs->ftype == FB_VCharacter)) {  // avoid quoting NULL or LOB handles
                *dest++ = '\'';
            }
			//crLOB
			if(bh != NULL) {
				fbcbhRelease(bh);
				bh = NULL;
			}
        }
        *dest = '\0';
    }

    if (dbis->debug >= 2) {
        PerlIO_printf(DBILOGFP, "dbd_st_execute: statement = >%s<\n", statement);
    }

    // execute statement
    imp_sth->result = PQexec(imp_dbh->conn, statement);

    // free statement string in case of input parameters
    if ((int)DBIc_NUM_PARAMS(imp_sth) > 0) {
        Safefree(statement);
    }

	//cr138 : tidied up
	if(imp_sth->result == NULL) {
		dbd_drv_error(sth, -1, "No result set metadata");
	} else if(fbcmdErrorsFound(imp_sth->result) == YES) {
		dbd_drv_error(sth, FBerrorCode(imp_dbh->conn, imp_sth->result), PQerrorMessage(imp_dbh->conn, imp_sth->result));
	} else if (fbcmdFetchHandle(imp_sth->result) != NULL) {
        // statement generating results
        num_fields = fbcmdColumnCount(imp_sth->result);
        imp_sth->current_row = 0;
        DBIc_NUM_FIELDS(imp_sth) = num_fields;
        DBIc_ACTIVE_on(imp_sth);
        ret = fbcmdRowCount(imp_sth->result);
        imp_sth->conn		= imp_dbh->conn;

        // It is possible that the user can reuse a statement handle to re-execute the statement and then fetch.  So, we need to test to see if we have a previous row handler and release it.
        if (imp_sth->rowHdlr != NULL) {
            fbcrhRelease(imp_sth->rowHdlr);
        }
        
        imp_sth->rawData	= NULL;	// not initialized, yet
        imp_sth->rowHdlr	= NULL; // not initialized, yet
		if (dbis->debug >= 2) {
			PerlIO_printf(DBILOGFP, "dbd_st_execute:  result set generated with %d rows\n", ret);
		}
	} else {
        ret = fbcmdRowCount(imp_sth->result);
		if (dbis->debug >= 2) {
			PerlIO_printf(DBILOGFP, "dbd_st_execute:  No result set generated with %d affected rows\n", ret);
		}
	}	
	//cr137 : now caters for warnings
	if(ret != -2 && fbcmdErrorCount(imp_sth->result) > 0) {
		char *errorClass = fbcmdErrorClass(imp_sth->result);
		if(errorClass[4] == '0' || errorClass[4] == '3')
			dbd_drv_warn(sth, PQerrorMessage(imp_dbh->conn, imp_sth->result));
	}
    // store the number of affected rows
    imp_sth->rows = ret;

	if (dbis->debug >= 2) {
        PerlIO_printf(DBILOGFP, "dbd_st_execute:  ret=%d\n", ret);
    }
	
    return ret;
}

// convert raw database value to perl value returns malloc'ed string
static char* getVal(imp_sth_t *imp_sth, void *col, int type, int* size)
{
	D_imp_dbh_from_sth;
    char	*value = NULL;

    switch(type)  {
        case FB_Boolean:
            switch(*(char*)col) {
				//cr137 : returns "FALSE"/"TRUE" when fb_bools is true (the default) and ""/"1" (perl bools) when false. 
                case 0:	
						if(imp_dbh->fb_bools)
							value = strdup("FALSE");
						else
							value = strdup("");			
						break;
                case 1:		
						if(imp_dbh->fb_bools)
							value = strdup("TRUE");
						else
							value = strdup("1");			
						break;
                default: 	value = strdup("NULL");		break;
            }
            break;

        case FB_YearMonth:
        case FB_Integer: {
            char	buf[100];
            sprintf(buf, "%d", *(int *)col);
            value	= strdup(buf);
            break;
        }

        case FB_SmallInteger:  {
            char	buf[100];
            sprintf(buf, "%d", *(short *)col);
            value	= strdup(buf);
            break;
        }

        case FB_DayTime:
        case FB_Real:
        case FB_Float:
        case FB_Decimal:	//cr136 Maybe decimal and numeric should be converted with the correct scale?
        case FB_Numeric: {
            char	buf[100];
            sprintf(buf, "%f", *(double *)col);
            value	= strdup(buf);
            break;
        }
        case FB_Bit:
        case FB_VBit:
            value = strdup("(no bit vectors)");
            /* FIXME: was obscure code:

            isBinary = 1;
            sv_setpvn(sv, "", 0);
            SvGROW(sv, size + 1);
            memcpy(((unsigned char *)SvPVX(sv)), val, size);
            SvCUR_set(sv, size);
            *SvEND(sv) = '\0';
            continue;
            */

        case FB_Date:
        case FB_Time:
        case FB_TimeTZ:
        case FB_Timestamp:
        case FB_TimestampTZ:
            value = strdup((char*)col);
            break;

        case FB_Character:
            value	= (char*)col;
            //cr137 : reinstated
            if(DBIc_has(imp_sth, DBIcf_ChopBlanks)) {
                int len   = strlen(value);
                char *str = value;
                while((len > 0) && (str[len-1] == ' ')) {
                    len--;
                }
                value[len] = '\0';
            }
            
        case FB_VCharacter:
            value	= strdup((char*)col);
            break;
		case FB_CLOB: { //crLOB
			char* handle = col;
			if(handle[0] == 0) {
				char* lob_handle = &handle[1];
				FBCBlobHandle* bh = fbcbhInitWithHandle(lob_handle);
				*size = 0;
				if(bh != NULL) {
					*size = fbcbhBlobSize(bh);
					if(*size > 0) {
						value = (char*)fbcdcReadCLOB(imp_dbh->conn, bh); //cr0135
						//ensure string is null terminated
						if(value != NULL && value[*size] != '\0') {
							char* value_nt = malloc(*size + 1);
							strncpy(value_nt, value, *size);
							value_nt[*size] = '\0';
							free(value);
							value = value_nt;
						}
					}
					fbcbhRelease(bh);
				}
				if(*size <= 0) {
					size = 0;
					value = strdup("");
				}
			}
			break;
		}
		case FB_BLOB: { //crLOB
			char* handle = col;
			if(handle[0] == 0) {
				char* lob_handle = &handle[1];
				FBCBlobHandle* bh = fbcbhInitWithHandle(lob_handle);
				*size = 0;
				if(bh != NULL) {
					value = fbcdcReadBLOB(imp_dbh->conn, bh);
					if(value != NULL) {
						*size = fbcbhBlobSize(bh);
					}
					fbcbhRelease(bh);
				}
			}
			break;
		}	

        default:
            value	= strdup("(unsupported type)");
            break;
    }
    return value;
}



AV *dbd_st_fetch(SV *sth, imp_sth_t *imp_sth)
{
    int		num_fields;
    int		i;

    AV*		av;
    void**	row;

    if (dbis->debug >= 1) {
        PerlIO_printf(DBILOGFP, "dbd_st_fetch\n");
    }

    // Check that execute() was executed sucessfully
    if (! DBIc_ACTIVE(imp_sth)) {
        dbd_drv_error(sth, 1, "no statement executing\n");
        return Nullav;
    }

    av = DBIS->get_fbav(imp_sth);
    num_fields	= AvFILL(av)+1;

    if(imp_sth->rawData == NULL || imp_sth->current_row >= (int)fbcrhRowCount(imp_sth->rowHdlr)) { //cr0135
        // fetch more rows
        imp_sth->rawData = fbcdcFetch(imp_sth->conn, ROWS_PER_FETCH, fbcmdFetchHandle(imp_sth->result));

        if (imp_sth->rawData == NULL)  {
            if (dbis->debug >= 1) {
                PerlIO_printf(DBILOGFP, "dbd_st_fetch: no raw data\n");
            }
            return Nullav;
        }
        // init row handler with new raw data
        if (imp_sth->rowHdlr != NULL) {
            fbcrhRelease(imp_sth->rowHdlr);
        }

        imp_sth->rowHdlr = fbcrhInitWith(imp_sth->rawData, imp_sth->result);
        imp_sth->current_row = 0;

        if (imp_sth->rowHdlr == NULL) {
            if (dbis->debug >= 1) {
                PerlIO_printf(DBILOGFP, "dbd_st_fetch:rowHdlr == NULL");
            }
            return Nullav;
        }
    } // if(imp_sth->rawData == NULL ...)

    if (dbis->debug >= 1) {
        PerlIO_printf(DBILOGFP, "dbd_st_fetch:about to fetch row %d with rh 0x%l08x\n", imp_sth->current_row+1, imp_sth->rowHdlr);
    }

    row	= fbcrhRowAtIndex(imp_sth->rowHdlr, imp_sth->current_row);

    if (row == NULL) {
        return Nullav;
    }
    
    // iterate over columns
    for(i = 0; i < num_fields; ++i) {
        void * col = row[i];
        SV *sv = AvARRAY(av)[i];
        int size = PQfsize(imp_sth->result, i);
        int type = PQftype(imp_sth->result, i);
        if (dbis->debug >= 1) {
            char *name = PQfname(imp_sth->result, i);
            PerlIO_printf(DBILOGFP, "dbd_st_fetch:%s, size %d, type %d\n", name == NULL ? "(NULL)": name, size, type);
        }
        if (col == NULL) {
            sv_setsv(sv, &sv_undef);
        } else {
            void *value = getVal(imp_sth, col, type, &size);
			if(type == FB_BLOB)
				sv_setpvn(sv, value, size);
			else 
				sv_setpv(sv, value);
            free(value);
        }
    }

    imp_sth->current_row++;
    if (dbis->debug >= 1) {
        PerlIO_printf(DBILOGFP, "end of dbd_st_fetch\n");
    }

    return av;
}

int dbd_st_blob_read(SV *sth, imp_sth_t *imp_sth, int field, long offset, long len, SV *destrv, long destoffset)
{
	D_imp_dbh_from_sth;
	int status = 0, is_null = 0;
	long retl = 0;
	SV *bufsv;
	char* handle = NULL;
	char* value = NULL;
	bufsv = SvRV(destrv);
	sv_setpvn(bufsv,"",0);      
	SvGROW(bufsv, (STRLEN)len+destoffset+1); //cr0135
	if(imp_sth->result != NULL && imp_sth->rowHdlr != NULL) {
		void** row	= fbcrhRowAtIndex(imp_sth->rowHdlr, imp_sth->current_row-1);
		if(row != NULL && field < (int)fbcmdColumnCount(imp_sth->result)) { //cr0135
			void* col = row[field];
			int type = PQftype(imp_sth->result, field);
			if(type == FB_BLOB) {
				char* handle = col;
				if(handle == NULL) {
					if (dbis->debug >= 1){
						PerlIO_printf(DBILOGFP, "dbd_st_blob_read:NULL value\n");
					}
					status = is_null = 1;
				}
				else if(handle[0] == 0) {
					char* lob_handle = &handle[1];
					FBCBlobHandle* bh = fbcbhInitWithHandle(lob_handle);
					if(bh != NULL) {
						value = fbcdcReadBLOB(imp_dbh->conn, bh);
						if(value != NULL) {
							status = 1;
							retl = fbcbhBlobSize(bh);
						}
						fbcbhRelease(bh);
					}
				}
				else if (dbis->debug >= 1){
					PerlIO_printf(DBILOGFP, "dbd_st_blob_read:not a valid BLOB handle\n");
				} 
			}
			else if (dbis->debug >= 1){
				PerlIO_printf(DBILOGFP, "dbd_st_blob_read:column is not a BLOB\n");
			} 
		} 
		else if (dbis->debug >= 1){
			PerlIO_printf(DBILOGFP, "dbd_st_blob_read:invalid row\n");
		}
	}
	else if (dbis->debug >= 1){
        PerlIO_printf(DBILOGFP, "dbd_st_blob_read:no result set\n");
    }
// where is value freed?	
	if(status) {
		if(is_null) {
			(void)SvOK_off(bufsv);
		}
		else {
			if(len > retl-offset) 
				len = retl-offset;
			if(len < 0)
				len = 0;
			memcpy(SvPVX(bufsv) + destoffset, value + offset, len);
			SvCUR_set(bufsv, destoffset+len);
			*SvEND(bufsv) = '\0';
		}
	}
	return status;
}


int dbd_st_rows(SV *sth, imp_sth_t *imp_sth)
{
    if (dbis->debug >= 1) {
        PerlIO_printf(DBILOGFP, "dbd_st_rows %d\n", imp_sth->rows);
    }

    return imp_sth->rows;
}

int dbd_st_finish(SV *sth, imp_sth_t *imp_sth)
{
    dTHR;
	D_imp_dbh_from_sth;
	
    if (dbis->debug >= 1) {
        PerlIO_printf(DBILOGFP, "dbd_st_finish\n");
    }

	//cr138 : added cancel fetch
	if (DBIc_ACTIVE(imp_sth)) {
		if (imp_sth->result) {
			char* fHandle = fbcmdFetchHandle(imp_sth->result);
			// clear the cursor handle (this needs to be done for manual commit mode)
			if(0 && fHandle != NULL) { // breaks table_info
				FBCMetaData *md = fbcdcCancelFetch(imp_dbh->conn, fbcmdFetchHandle(imp_sth->result));
				fbcmdRelease(md);
			}
			fbcmdRelease(imp_sth->result);
			imp_sth->result = 0;
			imp_sth->rows   = -1;	//cr137 was 0
		}
		
		if(imp_sth->rowHdlr) {
			fbcrhRelease(imp_sth->rowHdlr);
			imp_sth->rowHdlr = NULL;
		}
	}

    DBIc_ACTIVE_off(imp_sth);
    return 1;
}

//cr138 tidied up. Some of the clean-up code was redundant since dbd_st_finish is already called.
void dbd_st_destroy(SV *sth, imp_sth_t *imp_sth)
{
    if (dbis->debug >= 1) {
        PerlIO_printf(DBILOGFP, "dbd_st_destroy\n");
    }

    // Free off contents of imp_sth

    Safefree(imp_sth->statement);
	
    if (imp_sth->out_params_av)
        sv_free((SV*)imp_sth->out_params_av);

    if (imp_sth->all_params_hv) {
        HV *hv = imp_sth->all_params_hv;
        SV *sv;
        char *key;
        I32 retlen;
        hv_iterinit(hv);
        while( (sv = hv_iternextsv(hv, &key, &retlen)) != NULL ) {
            if (sv != &sv_undef) {
                phs_t *phs_tpl = (phs_t*)(void*)SvPVX(sv);
                sv_free(phs_tpl->sv);
            }
        }
        sv_free((SV*)imp_sth->all_params_hv);
    }
    DBIc_IMPSET_off(imp_sth); // let DBI know we've done it
}

int dbd_st_STORE_attrib(SV *sth, imp_sth_t *imp_sth, SV *keysv, SV *valuesv)
{
    if (dbis->debug >= 1) {
        PerlIO_printf(DBILOGFP, "dbd_st_STORE\n");
    }

    return FALSE;
}


static const char *fb_type_name(int type) {
    // FIXME: are these really FrontBase types/numbers?
    static const char *type_names[]	= {
        "PrimaryKey",// 0
        "boolean",	// 1
        "integer",	// 2
        "smallint",	// 3
        "float",	// 4
        "real",	// 5
        "double",	// 6
        "numeric",	// 7
        "decimal",	// 8
        "character",	// 9
        "varchar",	// 10
        "bit",	// 11
        "vbit",	// 12
        "date",	// 13
        "time",	// 14
        "timetz",	// 15
        "timestamp",// 16
        "timestamptz",// 17
        "yearmonth",// 18
        "daytime",	// 19
        "clob",	// 20
        "blob",	// 21
        "daytime"	// 22
    };
    
    const char*	name;
    
    if(type < 0 || type >= sizeof(type_names)/sizeof(*type_names)) {
        name = "undefined";
    } else {
        name = type_names[type];
    }
    
    return name;
}



SV *dbd_st_FETCH_attrib(SV *sth, imp_sth_t *imp_sth, SV *keysv)
{
    STRLEN kl;
    char *key = SvPV(keysv,kl);
    SV *retsv = Nullsv;
    int i;

    if (dbis->debug >= 1) { PerlIO_printf(DBILOGFP, "dbd_st_FETCH\n"); }

    if (! imp_sth->result) {
        return Nullsv;
    }

    if (kl == 4 && strEQ(key, "NAME")) {
        AV *av = newAV();
        retsv = newRV(sv_2mortal((SV*)av));
        for (i = 0; i < DBIc_NUM_FIELDS(imp_sth); i++) {
            av_store(av, i, newSVpv(PQfname(imp_sth->result, i),0));
        }
    } else if ( kl== 4 && strEQ(key, "TYPE")) {
        AV *av = newAV();
        retsv = newRV(sv_2mortal((SV*)av));
        for (i = 0; i < DBIc_NUM_FIELDS(imp_sth); i++) {
            av_store(av, i, newSViv(sql_type_from_fb_type(PQftype(imp_sth->result, i))));	//cr137
        }
    } else if (kl==9 && strEQ(key, "PRECISION")) {
        AV *av = newAV();
        retsv = newRV(sv_2mortal((SV*)av));
        for (i = 0; i < DBIc_NUM_FIELDS(imp_sth); i++) {
            av_store(av, i, newSViv(FBcolPrec(imp_sth, i)));	//cr137
        }
    } else if (kl==5 && strEQ(key, "SCALE")) {
		int scale = -1;
        AV *av = newAV();
        retsv = newRV(sv_2mortal((SV*)av));
        for (i = 0; i < DBIc_NUM_FIELDS(imp_sth); i++) {
			scale = FBcolScale(imp_sth->result, i);
            av_store(av, i, scale == -1 ? &sv_undef : newSViv(scale)); //cr137
        }
    } else if (kl==8 && strEQ(key, "NULLABLE")) {
        AV *av = newAV();
        retsv = newRV(sv_2mortal((SV*)av));
        for (i = 0; i < DBIc_NUM_FIELDS(imp_sth); i++) {
            av_store(av, i, newSViv(fbccmdIsNullable(fbcmdColumnMetaDataAtIndex(imp_sth->result, i)))); //cr137
        }
    } else if (kl==10 && strEQ(key, "CursorName")) {
        retsv = &sv_undef;
    } else if (kl==7 && strEQ(key, "fb_size")) {
        AV *av = newAV();
        retsv = newRV(sv_2mortal((SV*)av));
        for (i = 0; i < DBIc_NUM_FIELDS(imp_sth); i++) {
            av_store(av, i, newSViv(PQfsize(imp_sth->result, i)));
        }
    } else if (kl==7 && strEQ(key, "fb_type")) {
        AV *av = newAV();
        retsv = newRV(sv_2mortal((SV*)av));
        for (i = 0; i < DBIc_NUM_FIELDS(imp_sth); i++) {
            int		type	= PQftype(imp_sth->result, i);
            const char*	name 	= fb_type_name(type);
            av_store(av, i, newSVpv((char*)name, 0));
        }
    } else {
        return Nullsv;
    }

    return sv_2mortal(retsv);
}

//cr139
SV* fb_db_digestPassword(const char* password, const char* username)
{
	char digest[255];
	upper_string((char*)username);
	fbcDigestPassword(username, password, digest);
	return newSVpv(digest, 0);
}

void upper_string(char *s)
{
  while(*s != '\0') {
    *s = toupper((int)(*s));
    s++;
  }
}
