#!perl -w

# Create the "dbd_pg_test" table and the "dbd_pg_sequence" sequence.
# Because the table is used for the other tests, we bail out if we cannot create it.

use Test::More tests => 2;
use DBI qw(:sql_types);
use Sys::Hostname;
use strict;
$|=1;

# defaults for connection

my $dsn = $ENV{DBI_DSN};
my $user = "_SYSTEM";
my $passwd = "";
my $hostname = hostname;
my $dbname = "test";
my $dbpasswd = "";

# DSN_xxx values override defaults

if($dsn && $dsn =~ s/^(dbi):(\w*?)(?:\((.*?)\))?://i && lc($2) eq 'fb') {
	$dbpasswd = $1 if($3 && $3 =~ /\bfb_dbpasswd\b\s*=>\s*(\w+)/);
	$dbname = $1 if($dsn =~ /\bdbname\b\s*=\s*(\w+)/);
	$hostname = $1 if($dsn =~ /\bhost\b\s*=\s*([\.A-Za-z0-9_]+)/);
}
$user = $ENV{DBI_USER} if ($ENV{DBI_USER});
$passwd = $ENV{DBI_PASS} if ($ENV{DBI_PASS});

# FB_xxx values override DSN_xxx values

$hostname = $ENV{FB_HOST} if ($ENV{FB_HOST});
$dbname = $ENV{FB_DBNAME} if ($ENV{FB_DBNAME});
$dbpasswd = $ENV{FB_DBPASSWD} if ($ENV{FB_DBPASSWD});
$user = $ENV{FB_USER} if ($ENV{FB_USER});
$passwd = $ENV{FB_PASSWD} if ($ENV{FB_PASSWD});

my $dbh = DBI->connect("dbi:FB:dbname=$dbname;host=$hostname", $user, $passwd,
							{fb_dbpasswd => $dbpasswd, RaiseError => 1, PrintError => 0, AutoCommit => 1});
				
ok( defined $dbh, "Connect to database for object creation");

# Create the testing table
eval {
$dbh->do("DROP TABLE dbd_fb_test cascade");
};

my $SQL1 = qq{
	CREATE TABLE dbd_fb_test(
		id         integer not null primary key,
		lii        integer,
		pname      varchar(20) default 'Testing Default' ,
		val        varchar(255),
		score      float,
		Fixed      character(5),
		pdate      timestamp,
		"CaseTest" boolean,
		bytetest   bit(17)
)
};

my $SQL2 = qq{
	CREATE TABLE dbd_fb_datatypes(
		char12_				char(12),
		varchar20_			varchar(20),
		date_				date,
		time_				time,
		timestamp_			timestamp,
		timewithzone_		time with time zone,
		timestampwithzone_	timestamp with time zone,
		int4_				integer,
		decimal_			decimal(2,1),
		clob_				clob,
		blob_				blob,
		boolean_			boolean
)
};


eval {
	$dbh->do($SQL1);
	$dbh->do($SQL2);
};
if ($@) {
	if (! $DBI::errstr) {
		print STDOUT "Quitting tests! Could not create test table: $@\n";
	}
	else {
		print STDOUT "Quitting tests! Could not create test table: $DBI::errstr\n";
	}
	exit; # Force a hasty exit
}

ok( $dbh->disconnect(), 'Disconnect from database');
