#!/usr/local/bin/perl -w

# $Id: test.pl,v 1.3 2003/01/04 00:05:42 ryan Exp $

# Before `make install' is performed this script should be runnable with
# `make test'. After `make install' it should work as `perl test.pl'

######################### not tested explicitly
#
# AutoCommit
# commit
# rollback
# Active
# Statement
# attributes
# err
# pg_auto_escape
# quote
# type_info_all
#
######################### We start with some black magic to print on failure.


BEGIN { $| = 1; }
END {print "test failed\n" unless $loaded;}
use DBI ':sql_types';
use Sys::Hostname;
$loaded = 1;
use strict; # enforces lexical variable usage

######################### End of black magic.

my $hostname;  # lexical variable

if (defined($ENV{FRONTBASE_HOSTNAME})) {
	$hostname = $ENV{FRONTBASE_HOSTNAME};
}
else {
	$hostname = hostname;
}

if($hostname eq '') {
	$hostname = 'localhost';
}
	
my $user 	= "_SYSTEM";
my $passwd 	= "";
my $dbname 	= "test;host=$hostname";
my ($dbh0, $dbh, $sth);
my $table = "XXX";
my @row_ary;

sub croak
{
	my ($msg) = (@_); # @_ is parameter list, i.e. message passed to croak
	$sth->finish if (defined($sth));
	$dbh->disconnect if (defined($dbh));
	die($msg);
}

sub Check
{
	my ($fatal, $traceNext, $msg, $res) = (@_);
	my ($l) = (length($msg));
	my ($dots, $err) = ('.' x (40 - $l), $DBI::errstr);
	
	$fatal = 0 if(!defined($fatal));
	$err = "" if(!defined($err));

	if($res)
	{
		print "$msg$dots ok\n";
	}
	else
	{
 		print "$msg$dots not ok: $err\n" if(!$fatal);
 		croak "$msg$dots not ok: $err\n" if($fatal);
	}
	DBI->trace($traceNext);
}

DBI->trace(2); # make your choice
Check(1, 0, 'connect', $dbh = DBI->connect("dbi:FB:dbname=$dbname", $user, $passwd, { RaiseError => 0, AutoCommit => 0, PrintError => 1, fb_dbpasswd =>""}));

######################### create table
Check(0, 0, 'create table', $dbh->do("CREATE TABLE $table ( 
  char12_    char(12),
  varchar20_ varchar(20),
  date_      date,
  int4_      integer,
  decimal_      decimal(2,1)
  );"));

$sth = $dbh->prepare("SELECT UNIQUE FROM $table") || die('Can\'t get sequence');
$sth->execute();
my ($unique, $oldunique);
($oldunique)= $sth->fetchrow_array();

Check(0, 0, 'sequence first time', ($oldunique > 0 && $oldunique < 2147483647));
$sth->execute();
($unique)= $sth->fetchrow_array();

Check(0, 0, 'sequence next time', ($oldunique = $unique -1));
$sth->finish();

$sth = $dbh->table_info;
my @infos = $sth->fetchrow_array;
$sth->finish;

Check(0, 0, '$dbh->table_info', ( join(" ", @infos) =~ $table));
Check(0, 0, '$dbh->tables', ( join(" ", $dbh->tables) =~ $table));

######################### test various insert methods

# insert into table with $dbh->do($stmt)

Check(0, 0, 'insert plain', $dbh->do("INSERT INTO $table VALUES(
  '1',
  'Anjo Krank',
  DATE '1997-12-03',
  1234,
  1.1
  )"));

# insert into table with $dbh->prepare() with placeholders and $dbh->execute(@bind_values)

Check(0, 0, 'insert prepare', $sth = $dbh->prepare( "INSERT INTO $table 
  ( char12_, varchar20_, date_, int4_, decimal_ )
  VALUES ( ?, ?, DATE ?, ?, ? )
  " ) );

Check(0, 0, 'insert exec prepare',  $sth->execute (
  '2',
  'Edmund Mergl',
  '1997-12-03',
  1234,
  2.2
  ));

Check(0, 0, 'insert exec prepare',  $sth->execute (
  '3',
  'Geert B. Clemmensen',
  '1997-12-03',
  1234,
  3.0
   ));

# insert into table with $dbh->do($stmt, @bind_values)

Check(0, 0, 'insert bound values',  $dbh->do( "INSERT INTO $table 
  VALUES (?, ?, DATE ?, ?, ? )",
  {},
  '4',
  'Sandra Fritsche',
  '1997-12-03',
  1234,
  4.0
   ));

Check(0, 0, 'commit', $dbh->commit());
Check(0, 0, 'rollback', $dbh->rollback());
Check(0, 0, 'finish',  $sth->finish );

######################### test various select methods
Check(0, 0, 'update', $dbh->do("UPDATE $table SET int4_=1235 WHERE int4_ = 1234"));

# select from table using input parameters and and various fetchrow methods
Check(0, 0, 'select prepare with bind',  $sth = $dbh->prepare("SELECT * FROM $table where char12_  < ?"));
Check(0, 0, 'select bind', ( $sth->bind_param(1, '4', SQL_CHAR) )); 
Check(0, 0, 'select execute', ( $sth->execute )); 
while(@row_ary = $sth->fetchrow_array)
{
#	print  join(" ", @row_ary) . "\n";
}
Check(0, 0, 'finish',  $sth->finish );

Check(0, 0, 'select prepare with bind',  $sth = $dbh->prepare("SELECT * FROM $table where char12_  < ?"));
Check(0, 0, 'select bind', ( $sth->bind_param(1, '4', SQL_CHAR) )); 
Check(0, 0, 'select execute', ( $sth->execute )); 
Check(0, 0, 'finish',  $sth->finish );
#
#croak("\n");

# Check(0, 0, 'select simple',  $sth = $dbh->do("SELECT * FROM $table  where char12_ < '4'"));
# @row_ary = $sth->fetchrow_array;
# print  join(" ", @row_ary);

Check(0, 0, 'select prepare simple',  $sth = $dbh->prepare("SELECT * FROM $table  where char12_  < '4'"));
Check(0, 0, 'select execute',  $sth->execute());
@row_ary = $sth->fetchrow_array;
#print  join(" ", @row_ary);

Check(0, 0, 'select bind int prepare',  $sth = $dbh->prepare("SELECT * FROM $table  where int4_  < ?"));
Check(0, 0, 'select bind int bind',  $sth->bind_param(1, 10000, SQL_INTEGER));
Check(0, 0, 'select bind int exec', $sth->execute);

@row_ary = $sth->fetchrow_array;
#print  join(" ", @row_ary);
( join(" ", @row_ary) eq q{1 Anjo Krank 1997-12-03 1235 1.100000} ) 
    and print "\$sth->fetchrow_array ..... ok\n"
    or  print "\$sth->fetchrow_array ..... not ok: ", join(" ", @row_ary), "\n";

my $ary_ref = $sth->fetchrow_arrayref;
( join(" ", @$ary_ref) eq q{2 Edmund Mergl 1997-12-03 1235 2.200000} )
    and print "\$sth->fetchrow_arrayref .. ok\n"
    or  print "\$sth->fetchrow_arrayref .. not ok: ", join(" ", @$ary_ref), "\n";

my ($key, $val);
my $hash_ref = $sth->fetchrow_hashref;
( join(" ", (($key,$val) = each %$hash_ref)) eq q{VARCHAR20_ Geert B. Clemmensen} )
    and print "\$sth->fetchrow_hashref ... ok\n"
    or  print "\$sth->fetchrow_hashref ... not ok:  key = $key, val = $val\n";
# test various attributes

my @name = @{$sth->{NAME}};
( join(" ", @name) eq q{CHAR12_ VARCHAR20_ DATE_ INT4_ DECIMAL_} )
    and print "\$sth->{NAME} ............. ok\n"
    or  print "\$sth->{NAME} ............. not ok: ", join(" ", @name), "\n";

my @type = @{$sth->{TYPE}};
( join(" ", @type) eq q{9 10 13 2 8} )
    and print "\$sth->{TYPE} ............. ok\n"
    or  print "\$sth->{TYPE} ............. not ok: ", join(" ", @type), "\n";


my @fb_size = @{$sth->{fb_size}};
( join(" ", @fb_size) eq q{12 20 0 0 0} )
    and print "\$sth->{fb_size} .......... ok\n"
    or  print "\$sth->{fb_size} .......... not ok: ", join(" ", @fb_size), "\n";

my @fb_type = @{$sth->{fb_type}};
( join(" ", @fb_type) eq q{character varchar date integer decimal} )
    and print "\$sth->{fb_type} .......... ok\n"
    or  print "\$sth->{fb_type} .......... not ok: ", join(" ", @fb_type), "\n";

# test binding of output columns

$sth->execute or die $DBI::errstr;
my $gaga;
( $sth->bind_col(2, \$gaga) )
    and print "\$sth->bind_col ........... ok\n"
    or  print "\$sth->bind_col ........... not ok: ", $DBI::errstr;

$sth->fetch or die $DBI::errstr;
( $gaga eq q{Anjo Krank} )
    and print "\$sth->fetch .............. ok\n"
    or  print "\$sth->fetch .............. not ok: $gaga\n";

my ($char12, $vchar12, $date, $int4, $decimal);
( $sth->bind_columns(undef, \$char12, \$vchar12, \$date, \$int4, \$decimal) )
    and print "\$sth->bind_columns ....... ok\n"
    or  print "\$sth->bind_columns ....... not ok: ", $DBI::errstr;

$sth->fetch or die $DBI::errstr;
( "$char12, $vchar12, $date, $int4, $decimal" eq 
  q{2, Edmund Mergl, 1997-12-03, 1235, 2.200000} )
    and print "\$sth->fetch .............. ok\n"
    or  print "\$sth->fetch .............. not ok:  $char12, $vchar12, $date, $int4, $decimal\n";


$sth->finish or die $DBI::errstr;

# select from table using input parameters

$sth = $dbh->prepare( "SELECT * FROM $table where char12_ = ?" ) or die $DBI::errstr;

my $string = q{2};
$sth->bind_param(1, $string) or die $DBI::errstr;

# $dbh->{pg_auto_escape} = 1;
# is needed for $string above and is on by default
$sth->execute or die $DBI::errstr;

$sth->{ChopBlanks} = 1;
@row_ary = $sth->fetchrow_array;
( join(" ", @row_ary) eq q{2 Edmund Mergl 1997-12-03 1235 2.200000} ) 
    and print "\$sth->{ChopBlanks} ....... ok\n"
    or  print "\$sth->{ChopBlanks} ....... not ok: ", join(" ", @row_ary), "\n";

my $rows = $sth->rows;
( 1 == $rows )
    and print "\$sth->rows ............... ok\n"
    or  print "\$sth->rows ............... not ok: $rows\n";

$sth->finish or die $DBI::errstr;

######################### test blobs

#my $lobject = '/tmp/gaga';
#
#my $data = "testing large objects using blob_read";
#open(FD, ">$lobject") or die "can not open $lobject";
#print(FD $data);
#close(FD);
#
#$dbh->do("CREATE TABLE lobject ( id int4, loid oid )") or die $DBI::errstr;
#$dbh->do("INSERT INTO lobject (id, loid) VALUES (1, lo_import('$lobject') )") or die $DBI::errstr;
#
#unlink $lobject;
#
#$sth = $dbh->prepare("SELECT loid FROM lobject WHERE id = 1") or die $DBI::errstr;
#$sth->execute or die $DBI::errstr;
#my $lobj_id = $sth->fetchrow_array;
#my $blob = $sth->blob_read($lobj_id, 0, 0);
#( $data eq $blob )
#    and print "\$sth->blob_read .......... ok\n"
#    or  print "\$sth->blob_read .......... not ok: >$blob<\n";
#
#$sth->finish or die $DBI::errstr;

######################### disconnect and drop test database

# disconnect
Check(0, 0, 'drop table', $dbh->do("DROP TABLE $table CASCADE"));
$dbh->commit()  or print( $DBI::errstr);

( $dbh->disconnect )
    and print "\$dbh->disconnect ......... ok\n"
    or  die   "\$dbh->disconnect ......... not ok: ", $DBI::errstr;

# $dbh0->do("DROP DATABASE $dbname");
# $dbh0->disconnect;

print "test sequence finished.\n";

######################### EOF
